/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: NumericalDerivative.c,v 1.2 1996/05/20 18:10:37 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include <float.h>
#include "dmd/tools.h"

#define HDEFAULT 0.0001 		/* default fractional step size */
#define EPSILON (DBL_EPSILON * 1.e5)	/* smallest number to play with */

/* Take the numerical derivative of the function, cheaply.
 * INPUT:
 * fun(x, resources)	:= evaluates a function
 * res			:= resources to pass to fun
 * x			:= value at which to compute derivative
 * f			:= precomputed function value at x
 * tol			:= tolerance for error in derivative
 * OUTPUT:
 * dfdx			:= computed derivative
 *
 * Return non-zero if the derivative could not be evaluated.
 */
int NumericalDerivative(double (*fun)(double, void *), void *res,
	double x, double f, double tol, double *dfdx)
{
	double h = (x == 0.0 ? HDEFAULT : fabs(x * HDEFAULT));
	double flo, fhi;
	double err, result;

	if(h < EPSILON) h = HDEFAULT;

	do {
	  flo = fun(x - h, res);
	  fhi = fun(x + h, res);

	  result = (fhi - flo) / (2.0 * h);
	  err = (flo + result * h - f) / (f + EPSILON);
	  
	  if(fabs(err) <= tol) {
	    *dfdx = result;
	    return 0;
	  }

	  h *= 0.618;

	} while(h > EPSILON);

	*dfdx = 0.0;
	return -1;
}
