/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: amoeba.c,v 1.4 1996/05/29 21:11:16 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "tools.h"
#include "memory.h"

#define ALPHA 1.0	/* First attempt to move simplex */
#define BETA 0.5	/* Second attempt to move simplex */
#define GAMMA 2.0	/* Third attempt to move simplex */

/* Extrapolate by factor fac through the face of the simplex across
 * from the high point.  Try the new point, and replace the high point
 * if the new point is better.
 */
static int
reflect(double *result, double **p, double *y, double *psum, int ndim,
	int (*funk)(double *, double *, void *), void *res,
	int ihi, int *nfunk, double fac)
{
	int r, j;
	double fac1, fac2;
	double *ptry = dmalloc(ndim);

	fac1 = (1.0 - fac) / ndim;
	fac2 = fac1 - fac;
	for(j=0; j<ndim; ++j)
	  ptry[j] = psum[j] * fac1 - p[ihi][j] * fac2;

	++(*nfunk);
	if ((r = (*funk)(ptry, result, res)) == 0 && *result < y[ihi])
	{
		y[ihi]= *result;
		for (j=0; j<ndim; ++j)
		{
			psum[j] += ptry[j] - p[ihi][j];
			p[ihi][j] = ptry[j];
		}
	}
	myfree(ptry);
	return r;
}

/* Contract the simplex around vertex ilo.
 */
#define CONTRACT_RETRY 4
static int contract(int ilo, double **p, double *y, int ndim,
	int (*funk)(double *, double *, void *), void *res, int *nfunk)
{
	int i, j, try, r = 0;
	for(i=0; i<(ndim + 1); ++i)
	{
	  if (i != ilo)
	  {
	    for(try=0; try<CONTRACT_RETRY; ++try)
	    {
	      for(j=0; j<ndim; ++j)
		p[i][j] = 0.5 * (p[i][j] + p[ilo][j]);

	      ++(*nfunk);
	      if((r = (*funk)(p[i], y + i, res)) == 0) break;
	    }
	    if(r) return r;
	  }
	}
	return r;
}

struct local_sort_s {
	double *p;
	double y;
};

static int cmp(const void *a, const void *b)
{
	struct local_sort_s
	  *sa = (struct local_sort_s *) a,
	  *sb = (struct local_sort_s *) b;
	return sa->y < sb->y ? -1 :
	       sa->y > sb->y ? 1 : 0;
}

/* Downhill simplex method for minimization in multiple dimensions.
 * Adapted from _Numerical_Recipes_.
 *
 * p[i][j] = jth component of ith vertex, i=0 to ndim, j=0 to ndim-1.
 * y[i] = values of fun() at ith vertex.
 * ndim = number of parameters
 * ftol = fractional convergence of function value to minimum.
 * nmax = maximum allowed iterations (a few thousand is good).
 * funk(double p[], double *result, void *res) = function to be minimized;
 *   returns non-zero if p[] is an invalid set of parameters.  The 
 *   resources res are simply passed to funk.
 * nfunk = number of times that funk() was called.
 *
 * Upon return, the simplex vertices are ordered in increasing 
 * function value.
 *
 * This code can support multiple threads of control.
 */ 
int amoeba(double **p, double *y, int ndim, double ftol, int nmax,
	int (*funk)(double *, double *, void *), void *res, int *nfunk)
{
	int r = 0;
	int i, j, ilo, ihi, inhi;
	double ytry, ysave, rtol;
	double *psum;

	if(ndim < 3) {
	  dmd_errno = DMD_INVALID_ARGUMENT;
	  return -1;
	}

	psum = dmalloc(ndim);

	*nfunk = 0;

restart:
	for(j=0; j<ndim; ++j)
	  for(psum[j]=0.0, i=0; i<(ndim + 1); ++i) psum[j] += p[i][j];

	while(1) {
	  /* Determine the highest (worst), next-highest, and lowest
	   * point in the simplex.
	   */
	  ilo = 0;
	  ihi = y[0] > y[1] ? (inhi=1,0) : (inhi=0,1);
	  for (i=0; i<(ndim + 1); ++i)
	  {
		  if (y[i] < y[ilo]) ilo = i;

		  if (y[i] > y[ihi]) {
			  inhi = ihi;
			  ihi = i;
		  } else if (y[i] > y[inhi]) {
			  if (i != ihi) inhi=i;
		  }
	  }

	  /* Compute the fractional range from highest to lowest.
	   * Return if we are done.
	   */
	  rtol = 2.0*fabs(y[ihi]-y[ilo])/(fabs(y[ihi])+fabs(y[ilo]));
	  if (rtol < ftol) goto done;
	  if (*nfunk >= nmax) {r = -1; goto done;}

	  /* A new iteration begins here.
	   * Reflect the highest point across the simplex face.
	   */
	  r = reflect(&ytry, p, y, psum, ndim, funk, res, ihi, nfunk, -ALPHA);
	  if (r == 0 && ytry <= y[ilo])
	  {
	    /* Since we did well, try reflecting this one again.
	     */
	    r = reflect(&ytry, p, y, psum, ndim, funk, res, ihi, nfunk, GAMMA);
	  }

	  /* If we didn't do very well, consider contracting.
	   * At first, combine a contraction with a reflection, 
	   * but then, just contract in a last ditch effort.
	   * After a pure contraction, we shall have to restart.
	   */
	  if (r || ytry >= y[inhi])
	  {
	    ysave = y[ihi];
	    r = reflect(&ytry, p, y, psum, ndim, funk, res, ihi, nfunk, BETA);
	    if (r || ytry >= ysave)
	    {
	      /* We still can't get rid of the highest point, so 
	       * contract again around the lowest point and restart
	       * the algorithm with the resulting simplex.
	       */
	      if((r = contract(ilo, p, y, ndim, funk, res, nfunk))) goto done;
	      goto restart;
	    }
	  }
	} /* 1 (cycle loop) */

done:
	myfree(psum);

	/* Order the simplex vertices by increasing function
	 * value.
	 */
	{
	  struct local_sort_s 
	    *srt = mymalloc((ndim + 1) * sizeof(struct local_sort_s));
	  for(i=0; i<(ndim + 1); ++i) {
	    srt[i].p = p[i];
	    srt[i].y = y[i];
	  }
	  qsort(srt, ndim + 1, sizeof(struct local_sort_s), cmp);
	  for(i=0; i<(ndim + 1); ++i) {
	    p[i] = srt[i].p;
	    y[i] = srt[i].y;
	  }
	  myfree(srt);
	}

	return r;
}

