/* Convert ANSI/IEEE format arithmetic numbers to/from the local format.
 * Copyright (c) 1995 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: ansiconvert.c,v 1.3 1995/09/06 16:38:37 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "ansiconvert.h"

#define W32 4294967296.0 /* 2^32 */
#define W23 8388608.0    /* 2^23 */
#define W20 1048576.0    /* 2^20 */

/* Floating point numbers are stored in ANSI/IEEE as follows:
 *
 * s = sign bit, e = biased exponent, f = fraction
 *
 * 4-byte float:
 * Byte no.
 * ...0...|...1...|...2...|...3...|
 * seeeeeeeefffffffffffffffffffffff
 *            
 * 8-byte float:
 * Byte no.
 * ...0...|...1...|...2...|...3...|...4...|...5...|...6...|...7...|
 * seeeeeeeeeeeffffffffffffffffffffffffffffffffffffffffffffffffffff
 *           
 */

/* Convert a 4 byte buffer containing an ANSI IEEE format floating
 * point number into the local machine representation.
 * IEEE special values are not supported. XXX
 */
double ansifloat(unsigned char *data)
{
  unsigned int e = ((data[0] & 0x7f) << 1) | ((data[1] & 0x80) >> 7);
  if (e==0) return 0.0;
  else {
    unsigned char s = data[0] & 0x80;
    unsigned long flo;
    int ex = e - 127;
    double f; 
    flo = data[3] | (data[2] << 8) | ((data[1] & 0x7f) << 16);
    f = 1.0 + flo/W23;
    f = ldexp(f, ex);
    if (s) f *= -1.0;
    return f;
  }
}

/* Convert an 8 byte buffer containing an ANSI IEEE format floating
 * point number into the local machine representation.
 * IEEE special values are not supported. XXX
 */
double ansidouble(unsigned char *data)
{
  unsigned int e = ((data[0] & 0x7f) << 4) | ((data[1] & 0xf0) >> 4);
  if (e==0) return 0.0;
  else {
    unsigned char s = data[0] & 0x80;
    unsigned long flo, fhi;
    int ex = e - 1023;
    double f; 
    flo = data[7] | (data[6] << 8) | (data[5] << 16) | (data[4] << 24);
    fhi = data[3] | (data[2] << 8) | ((data[1] & 0x0f) << 16);
    f = 1.0 + (fhi + flo/W32)/W20;
    f = ldexp(f, ex);
    if (s) f *= -1.0;
    return f;
  }
}

/* Convert a double into ANSI/IEEE 4-byte format.  Use the supplied
 * buffer, or return a pointer to a static internal buffer.
 * IEEE special values are not supported. XXX
 */
unsigned char *floatansi(unsigned char *data, double f)
{
  static unsigned char buf[4];
  int e;
  unsigned char s = 0;
  unsigned long flo;

  if(data == NULL) data = buf;

  if(f == 0.0) {
    e = 0;
    flo = 0;
  } else {
    if(f < 0.0) { s = 0x80; f = -f; }
    f = frexp(f, &e);
    e = e + 126;
    f = W23 * (2.0 * f - 1.0);
    flo = f;
  }

  data[0] = s | ((e & 0x1ff) >> 1);
  data[1] = ((e & 0x1) << 7) | ((flo & 0x7f0000) >> 16);
  data[2] = (flo & 0xff00) >> 8;
  data[3] = flo & 0xff;

  return data;
}

/* Convert a double into ANSI/IEEE 8-byte format.  Use the supplied
 * buffer, or return a pointer to a static internal buffer.
 * IEEE special values are not supported. XXX
 */
unsigned char *doubleansi(unsigned char *data, double f)
{
  static unsigned char buf[8];
  int e;
  unsigned char s = 0;
  unsigned long flo, fhi;

  if(data == NULL) data = buf;

  if(f == 0.0) {
    e = 0;
    flo = fhi = 0;
  } else {
    if(f < 0.0) { s = 0x80; f = -f; }
    f = frexp(f, &e);
    e = e + 1022;
    f = W20 * (2.0 * f - 1.0);
    fhi = f;
    flo = W32 * (f - fhi);
  }

  data[0] = s | ((e & 0x7f0) >> 4);
  data[1] = ((e & 0xf) << 4) | ((fhi & 0x0f0000) >> 16);
  data[2] = (fhi & 0xff00) >> 8;
  data[3] = fhi & 0xff;
  data[4] = (flo & 0xff000000) >> 24;
  data[5] = (flo & 0xff0000) >> 16;
  data[6] = (flo & 0xff00) >> 8;
  data[7] = flo & 0xff;

  return data;
}

/* Convert a 4 byte buffer containing an ANSI IEEE format integer
 * into the local machine representation.
 */
int ansiint(unsigned char *data)
{
  unsigned int i = data[3] | (data[2] << 8) | (data[1] << 16) | (data[0] << 24);
  return (int) i;
}

/* Convert an integer on the local machine into an ANSI/IEEE 4
 * byte integer.
 */
unsigned char *intansi(unsigned char *data, int i)
{
  static unsigned char buf[4];
  unsigned int u = i;

  if(data == NULL) data = buf;

  data[0] = (u & 0xff000000) >> 24;
  data[1] = (u & 0xff0000) >> 16;
  data[2] = (u & 0xff00) >> 8;
  data[3] = (u & 0xff);

  return data;
}

