/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: calcfermi.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

/* Compute Fermi-Dirac distribution for a supplied list of eigenvalues
 * and, optionally, fractional weights.
 *
 * Translated from Jamie Morris's Fortran 77 by f2c(1), 6 April 1994.
 * Modified by David Deaven (DMD) <deaven@ishmael.ameslab.gov>
 * 12 July 1994 -- fix indexing problems causing access violation (DMD)
 * 22 July 1994 -- harden floating point ops for non-IEEE machines,
 *   and make the bisection code more robust. (DMD)
 *
 * Output:
 *   chem=chemical potential
 *   selec=electron entropy
 *   focc(i)=occupancy of energy level i
 * Input:
 *   chem=guessed chemical potential, OK if totally stupid
 *   eigen(i)=energy of level i (sorted, eigen(i)<=eigen(i+1))
 *   weight(i)=weight of level i (number of states, fractional OK)
 *     if weight==NULL, unit weights are supplied.
 *   telec=electron temperature
 *     if telec=0.0 then the eigenvalues must be sorted into ascending order
 *   norb=number of calculated eigenvectors
 *   ns=# of occupied states, i.e. sum(focc) upon return.
 *
 */
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "tools.h"

int calcfermi_dumped; /* XXX */

#define MAXIT 256 /* maximum iterations */
#define MAXNIT 32 /* maximum "nice" iterations (non-bisecting) */
#define EPSILON 1.e-10
void calcfermi(double *chem, double *selec, double focc[], double eigen[],
	double weight[], double telec, int norb, int ns)
{
    double elec, diff, deri, vexp=0.0, w, f1, f2;
    double emin, emax;
    int i, m;

    calcfermi_dumped = 0; /* XXX */

    if (telec > 0.0)
    {
	emin = emax = eigen[0];
	for(i=1; i<norb; ++i)
	{
	    if(eigen[i] < emin) emin = eigen[i];
	    if(eigen[i] > emax) emax = eigen[i];
	}
	for (m=0; ; ++m) /* iteration */
	{
	    elec = deri = (*selec) = 0.0;
	    for (i=0; i<norb; ++i)
	    {
		w = (weight==NULL ? 1.0 : weight[i]);
		vexp = (eigen[i] - (*chem)) / telec;
		if (vexp < 100.0)
		{
		    vexp = exp(vexp);
		    focc[i] = 1.0 / (vexp + 1.0);
		    deri += w * vexp * focc[i] * focc[i] / telec;
		}
		else
		    focc[i] = 0.0;

		elec += w*focc[i];
		if (focc[i] > DBL_EPSILON && focc[i] < (1.0 - DBL_EPSILON))
		{
		    f1 = focc[i];
		    f2 = 1.0 - focc[i];
		    (*selec) -= 2.0 * w * (f1*log(f1) + f2*log(f2));
		}
	    }
	    diff = elec - (double) ns;
	    if (fabs(diff) < EPSILON) break;

	    /* Do bisection method if half our time is up.  This
	     * is a failsafe way to home in on the answer, but
	     * it converges more slowly.
	     */
	    if(m == MAXNIT)
	      *chem = 0.5 * (emin + emax);
	    else if(m > MAXNIT)
	    {

	      if(diff < 0.0)
	        emin = *chem;
	      else
	        emax = *chem;
	      *chem = 0.5 * (emin + emax);

	    }
	    else /* use a fancier method */
	    {

	      /* Set vexp to desired step.  Watch for a
	       * terrace in f(x), and respond with a fixed step size.
	       */
	      if (2.0*deri*(emax-emin) < fabs(diff)*MAXIT)
		vexp = - (emax-emin) * diff / (double) norb; /* terrace */
	      else
	        vexp = - diff / deri; /* Newton's method */

	      if(fabs(vexp) > (emax-emin)) /* step is too much */
		vexp = copysign((emax-emin)/(double) m, vexp);

	      (*chem) += vexp;
	    }

	    if (m >= MAXIT)
	    {
	      fprintf(stderr,
		"calcfermi: did not converge: chem=%g diff=%g\n", *chem, diff);
	      calcfermi_dumped = 1; /* XXX */
	      *chem = *selec = 0.0;
	      for (i=0; i<norb; ++i)
		  focc[i] = (i < ns ? 1.0 : 0.0);
	      return; /* XXX */
	    }
	    else if (m > (MAXIT-8)) /* trace the last few steps */
	    {
	      fprintf(stderr,
		"calcfermi: m=%d: de=%g deri=%g chem=%g diff=%g step=%g\n",
		m, emax-emin, deri, *chem, diff, vexp);
	    }
	}
    }
    else
    {
	*selec = 0.0;
	*chem = eigen[ns];
	for (i=0; i<norb; ++i)
	    focc[i] = (eigen[i] <= (*chem) ? 1.0 : 0.0);
    }
}
#undef MAXIT
#undef MAXNIT
