/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: conjgrad.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

#include <math.h>
#include <stdio.h>
#include "memory.h"
#include "tools.h"

#define ZEPS 1.e-20 /* effective machine accuracy */

/* These control the default settings.
 */
#define LINTOL   1.e-4 /* fractional tolerance for line minimization */
#define LINITER  8     /* max iterations for line minimization */
#define FLETCHER 0     /* use Fletcher-Reeves instead of Polak-Ribiere */

/* Conjugate gradient optimization.
 * 
 * Input:
 *   x[] = coordinates
 *   n = number of coordinates.
 *   ftol = fractional tolerance for gradient.
 *   iter = maximum number of line minimizations
 *   fun(double x[], f[], *ep, int n) = function to calculate
 *     function and -gradient (forces)
 *   fout = file to write info to, if NULL then be quiet.
 * 
 * Output:
 *   x[] = coordinates.
 *   f[] = forces.
 *   fret = minimized function value.
 *
 *   return value = -1 for too many iterations, or a non-zero
 *     return code from fun().  If such an error occurs, the 
 *     parameters are set to the last known valid set.
 *
 * This routine supports multiple threads of control.
 *
 * David Deaven deaven@ishmael.ameslab.gov, starting with numerical recipes.
 */
int conjgrad(double x[], double f[], double *fret, int n, double ftol,
    int iter, int (*fun)(double *, double *, double *, int), FILE *fout)
{
    return conjgrad_engine(x, f, fret, n, ftol, iter, 
      FLETCHER, LINTOL, LINITER, fun, fout);
}

/* In addition to the arguments to the above, this call allows
 * specification of the line minimization parameters.
 */
int conjgrad_engine(double x[], double f[], double *fret, int n,
    double ftol, int iter, int fletcher, double lintol, int liniter,
    int (*fun)(double *, double *, double *, int), FILE *fout)
{
    int j, its, r;
    double *g, *h, gg, fp, dgg, gam, dot, frac;
    double *xsave;

    g = dmalloc(n);
    h = dmalloc(n);
    xsave = dmalloc(n);

    if((r=(*fun)(x, f, &fp, n))) goto done; /* also inits r */

    if(fout) fprintf(fout, "%3s %12s %20s %12s\n",
       "it", "dot", "E", "dE/|E|");

    for (j = 0; j < n; ++j) h[j] = g[j] = f[j]; dot=1.0;
    if(fout) fprintf(fout, "%3d %12.4e %20.12e\n", 0, dot, fp);

    for (its=1; its<=iter; ++its)
    {
	for(j=0; j<n; ++j) xsave[j] = x[j];
	if((r=linmin(x, h, n, lintol, liniter, fun, fret, f))) {
	  for(j=0; j<n; ++j) x[j] = xsave[j];
	  (*fun)(x, f, fret, n);
	  goto done;
	}
	frac = fabs(*fret - fp) / (fabs(fp) + ZEPS);
	if(fout)
	  fprintf(fout, "%3d %12.4e %20.12e %12.4e\n", its, dot, *fret, frac);
	if (frac <= ftol) goto done;
	fp = *fret;

	/* The original code called fun() here; but my linmin()
	 * takes care to return after having called it at the
	 * final point, so we can omit one call.
	 */
	/* if((r=(*fun)(x, f, &fp, n))) goto done; */
	gg = dgg = 0.0;
	for (j = 0; j < n; ++j)
	{
	    gg += g[j] * g[j];
	    if(fletcher)
	      dgg += f[j] * f[j];
	    else
	      dgg += (f[j] - g[j]) * f[j];
	}
	if (gg == 0.0) goto done; /* unlikely: gradient == zero */
	gam = dgg / gg;
	dot = 0.0;
	for (j = 0; j < n; ++j)
	{
	    g[j] = f[j];
	    h[j] = g[j] + gam * h[j];
	    dot += h[j] * f[j];
	}
    }
    if(fout) fprintf(fout, "conjgrad: maximum iterations exceeded\n");
    r = -1;

done:
    myfree(xsave);
    myfree(g);
    myfree(h);
    return r;
} 
