/* A small set of routines to aid in timing. 
 * Copyright (c) 1995 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: dmdtimer2.c,v 1.3 1995/10/05 16:06:00 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <string.h>
#include "memory.h"
#include "dmdtimer.h"

struct timedblock {
	char *id;		/* name of block */
	double time;		/* total time spent within block */
	double tchild;		/* total time spent in children */
	int call;		/* number of times block was entered */
};

static struct timedblock tb_zero = {"__start__", 0.0, 0.0, 1};
static struct timedblock **tbbase = NULL;
static unsigned int ntb = 0;

#define NFL 8			/* how many significant file name chars */
struct stackentry {
	struct timedblock *tb;	/* which block */
	double t_enter;		/* when did entry occur */
	char file[NFL];		/* filename */
	int line;		/* line number */
};

static struct stackentry *stack = NULL;
static int nstack = 0, maxstack = 0;

static struct timedblock *newtb(char *id)
{
	struct timedblock *new = mymalloc(sizeof(struct timedblock));
	new->id = mystrdup(id);
	new->time = 0.0;
	new->tchild = 0.0;
	new->call = 0;
	tbbase = myrealloc(tbbase, (ntb + 1) * sizeof(struct timedblock *));
	tbbase[ntb++] = new;
	return new;
}

static struct timedblock *findtb(char *id)
{
	int i;
	for(i=0; i<ntb; ++i)
	  if(strcmp(id, tbbase[i]->id) == 0) return tbbase[i];

	return NULL;
}

/* Enter a new timed block.
 */
void timerenter(char *id, char *file, int line)
{
	struct timedblock *tb;
	double t = MyClock();
	static int warn = 32;

	/* Initialize the list of timed blocks, and the stack,
	 * if it hasn't been done yet.
	 */
	if(ntb == 0) {
	  tbbase = myrealloc(tbbase, (ntb = 1) * sizeof(struct timedblock *));
	  tbbase[0] = &tb_zero;
	  stack = myrealloc(stack, (maxstack = 16) * sizeof(struct stackentry));
	  stack[0].tb = tbbase[0];
	  stack[0].t_enter = 0.0;
	  strncpy(stack[0].file, file, NFL);
	  stack[0].line = line;
	  nstack = 1;
	}

	if((nstack + 1) > maxstack) {
	  maxstack += 16;
	  stack = myrealloc(stack, maxstack * sizeof(struct stackentry));
	}

	if((tb = findtb(id)) == NULL) tb = newtb(id);

	++(tb->call);

	/* We just entered */
	stack[nstack].tb = tb;
	stack[nstack].t_enter = t;
	stack[nstack].line = line;
	strncpy(stack[nstack].file, file, NFL);

	if(++nstack > warn) {
	  fprintf(stderr, 
	    "warning: dmdtimer stack nested %d levels\n", nstack);
	  warn += 32;
	}
}

/* Leave a timed block, reverting to the previous context.
 */
void timerleave(char *file, int line)
{
	double dt, t = MyClock();

	if(nstack < 2) {
	  fprintf(stderr, "timerleave(): bad nesting, too many LEAVE()s\n");
	  return;
	}

	--nstack;

	if(strncmp(stack[nstack].file, file, NFL)  != 0) {
	  fprintf(stderr, "timerleave(): bad nesting, LEAVE() in "
	    "file `%s', line %d matches ENTER in file `%s', line %d\n",
	    file, line, stack[nstack].file, stack[nstack].line);
	}

	dt = t - stack[nstack].t_enter;

	/* the leaving block gets some added time */
	stack[nstack].tb->time += dt;

	/* the parent's child time also gets it */
	stack[nstack - 1].tb->tchild += dt;
}

#if 0
static int cmp(const void *a, const void *b)
{
	struct timedblock **tba = a, **tbb = b;
	double ta = (*tba)->time - (*tba)->tchild;
	double tb = (*tbb)->time - (*tbb)->tchild;
	return (ta > tb ? -1 : ta == tb? 0 : 1); 
}
#else
static int cmp(const struct timedblock **a,
               const struct timedblock **b)
{
        double ta = (*a)->time - (*a)->tchild;
        double tb = (*b)->time - (*b)->tchild;
        return (ta > tb ? -1 : ta == tb? 0 : 1);
}
#endif

void timerreport(FILE *f)
{
	int i;
	double dt, total;
	struct timedblock *tb;

	if(nstack != 1) {
	  fprintf(f,
	    "(REPORT() called before %d timed blocks completed)\n",
	    nstack - 1);
	}

	qsort(tbbase, ntb, sizeof(struct timedblock *),
              (int (*)(const void *,const void *)) cmp);

	for(total=0.0, i=0; i<ntb; ++i)
	{
	  if((tb = tbbase[i]) != &tb_zero)
	    total += tb->time - tb->tchild;
	}

	fprintf(f, "  time/sec(frac) +children/sec    calls  name\n");
	for(i=0; i<ntb; ++i)
	{
	  if((tb = tbbase[i]) != &tb_zero)
	  {
	    dt = tb->time - tb->tchild;
	    fprintf(f, "%10.3f(%4.2f) %13.3f %8d  %s\n", 
	      dt, dt / total, tb->time, tb->call, tb->id);
	  }
	}
	fprintf(f, "%10.3f total\n", tb_zero.tchild);
}

