/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: eistql2.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

/* The following code has been translated from FORTRAN to C using 
 * f2c(), and subsequently modified to optimize on a variety of
 * systems.  The detected compiler defines and the results are 
 * listed below:
 *
 * BLAS (any system)
 *    Use the BLAS library routines
 *
 * _AIX  (IBM RS/6000)
 *    #pragma for non-overlapping pointers
 *
 * _I860_ (Intel Paragon, OSF or SUNMOS)
 *    #pragma for non-overlapping pointers
 *    built in math functions
 *    BLAS calls
 *
 * David Deaven <deaven@ishmael.ameslab.gov>
 */

/* This subroutine is a translation of the algol procedure tql2,
 * Num. Math. 11, 293-306(1968) by Bowdler, Martin, Reinsch, and
 * Wilkinson.
 * Handbook for Auto. Comp., Vol.II-Linear Algebra, 227-240(1971).
 *
 * this subroutine finds the eigenvalues and eigenvectors
 * of a symmetric tridiagonal matrix by the ql method.
 * the eigenvectors of a full symmetric matrix can also
 * be found if  tred2  has been used to reduce this 
 * full matrix to tridiagonal form. 
 *
 * On input 
 *
 *    nm must be set to the row dimension of two-dimensional 
 *      array parameters as declared in the calling program 
 *      dimension statement. 
 *
 *    n is the order of the matrix. 
 *
 *    d contains the diagonal elements of the input matrix. 
 *
 *    e contains the subdiagonal elements of the input matrix 
 *      in its last n-1 positions.  e(1) is arbitrary. 
 *
 *    z contains the transformation matrix produced in the 
 *      reduction by  tred2, if performed.  if the eigenvectors 
 *      of the tridiagonal matrix are desired, z must contain 
 *      the identity matrix. 
 *
 * On output 
 *
 *    d contains the eigenvalues in ascending order.  if an 
 *      error exit is made, the eigenvalues are correct but 
 *      unordered for indices 1,2,...,ierr-1. 
 *
 *    e has been destroyed. 
 *
 *    z contains orthonormal eigenvectors of the symmetric 
 *      tridiagonal (or full) matrix.  if an error exit is made, 
 *      z contains the eigenvectors associated with the stored 
 *      eigenvalues. 
 *
 *    ierr is set to 
 *      zero       for normal return, 
 *      j          if the j-th eigenvalue has not been 
 *                     determined after 30 iterations. 
 *
 * Questions and comments should be directed to Burton S. Garbow, 
 * Mathematics and Computer Science div, Argonne National Laboratory 
 * This version dated August 1983.
 */

#include <math.h>
#include "eisdiag.h"

#ifdef _I860_
#define sqrt(x) __builtin_sqrt(x)
#define fabs(x) __builtin_fabs(x)
#ifndef BLAS
#define BLAS
#endif
#endif

#define PYTHAG(a,b) (sqrt(a*a+b*b))

int eistql2(double *d, double *e, int n, double *z)
{
    double c, f, g, h;
    int i, j, k, l, m;
    double p, r, s, c2, c3=0.0, s2=0.0;
    int l1, l2;
    int ii;
    double dl1, el1;
    int mml;
    double tst1, tst2;

#ifdef _AIX
#pragma disjoint(*z,*d,*e)
#endif

#ifdef _I860_
#pragma routine safe 
#endif

    if (n == 1) /* trivial case */
    {
    	return 0;
    }

    for (i=1; i<n; ++i) e[i-1] = e[i];

    f = 0.;
    tst1 = 0.;
    e[n-1] = 0.;

    for (l=0; l<n; ++l)
    {
	j = 0;
	h = fabs(d[l]) + fabs(e[l]);
	if (tst1 < h) tst1 = h;

        /* look for small sub-diagonal element.
	 * e(n) is always zero, so there is no exit
	 * through the bottom of the loop.
	 */
	for (m=l; m<n; ++m)
	{
	    tst2 = tst1 + fabs(e[m]);
	    if (tst2 == tst1) break;
	}

	if (m != l)
	{
	  do {
	    if (j == 30) /* too many iterations */
	    { 
	      return 1;
	    }
	    ++j;
	    /* form shift */
	    l1 = l + 1;
	    l2 = l1 + 1;
	    g = d[l];
	    p = (d[l1] - g) / (e[l] * 2.0);
	    r = PYTHAG(p, 1.0);
	    r = p + copysign(r, p);
	    d[l] = e[l] / r;
	    d[l1] = e[l] * r;
	    dl1 = d[l1];
	    h = g - d[l];
	    if (l2 <= n-1)
	    {
	      for (i=l2; i<n; ++i) d[i] -= h;
	    }

	    f += h;
	    /* ql transformation */
	    p = d[m];
	    c = 1.0;
	    c2 = c;
	    el1 = e[l1];
	    s = 0.0;
	    mml = m - l;
	    /* for i=m-1 step -1 until l do --
	     */
	    for (ii=1; ii<=mml; ++ii)
	    {
		c3 = c2;
		c2 = c;
		s2 = s;
		i = m - ii;
		g = c * e[i];
		h = c * p;
		r = PYTHAG(p, e[i]);
		e[i + 1] = s * r;
		s = e[i] / r;
		c = p / r;
		p = c * d[i] - s * g;
		d[i + 1] = h + s * (c * g + s * d[i]);
#ifdef BLAS
		drot(n, z + (i+1)*n, 1, z + i*n, 1, c, s);
#else
		/* form vector */
		for (k = 0; k < n; ++k)
		{
		    h = z[(i+1)*n +k];
		    z[(i+1)*n+k] = s * z[i*n+k] + c * h;
		    z[i*n+k] = c * z[i*n+k] - s * h;
		}
#endif
	    }

	    p = -s * s2 * c3 * el1 * e[l] / dl1;
	    e[l] = s * p;
	    d[l] = c * p;
	    tst2 = tst1 +  fabs(e[l]);
	  } while (tst2 > tst1);
	} /* m != l */

	d[l] += f;
    }

    /* Order eigenvalues and eigenvectors
     */
    for (ii = 1; ii <n; ++ii)
    {
	i = ii - 1;
	k = i;
	p = d[i];

	for (j = ii; j < n; ++j)
	{
	    if (d[j] < p)
	    {
	      k = j;
	      p = d[j];
	    }
	}

	if (k != i)
	{
	  d[k] = d[i];
	  d[i] = p;
	  for (j = 0; j < n; ++j)
	  {
	      p = z[i*n+j];
	      z[i*n+j] = z[k*n+j];
	      z[k*n+j] = p;
	  }
	}
    }
    return 0;
}
