/* Information about the periodic table of the elements
 * This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: element.c,v 1.4 1996/08/22 23:17:35 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "element.h"
#include "memory.h"

/* All numbers are in atomic units (au): amu, bohr
 * My first guess at the neutral radius (in bohr) is half the
 * bond length (in angstroms) taken from the 
 * _Handbook_of_Chemistry_and_Physics_ (Table F-181), except
 * in the case of the noble gases, where I use a steadily increasing
 * size as the mass increases.
 *
 *                     Atomic Average  Neutral  Bonding
 * Name  Fullname      Number  Mass     Radius   Radius  Color
 */
static Element table[] = {
  {"?",  "generic",      0,  1.0,       0.100,    0.00, "black"},
  {"Q",  "point charge", 0,  1.0,       0.000,    0.00, "red"},
  {"H",  "hydrogen",     1,  1.00797,   0.370,    1.20, "snow"},
  {"He", "helium",       2,  4.00260,   0.300,    1.20, "ivory"},
  {"Li", "lithium",      3,  6.94,      1.520,    1.20, "azure"},
  {"Be", "beryllium",    4,  9.01218,   1.113,    1.20, "lavender"},
  {"B",  "boron",        5,  10.811,    0.790,    1.55, "gray"},
  {"C",  "carbon",       6,  12.01115,  0.750,    1.55, "tan"},
  {"N",  "nitrogen",     7,  14.0067,   0.550,    1.55, "blue"},
  {"O",  "oxygen",       8,  15.9994,   0.604,    1.55, "SkyBlue"},
  {"F",  "florine",      9,  18.9984,   0.709,    1.55, "green"},
  {"Ne", "neon",        10,  20.17,     0.400,    1.55, "red"},
  {"Na", "sodium",      11,  22.98977,  1.858,    1.55, "yellow"},
  {"Mg", "magnesium",   12,  24.305,    1.599,    1.55, "silver"},
  {"Al", "aluminum",    13,  26.9815,   1.432,    2.90, "gray"},
  {"Si", "silicon",     14,  28.086,    1.176,    2.60, "SteelBlue"},
  {"P",  "phosphorus",  15,  30.9738,   1.105,    2.60, "red"},
  {"S",  "sulfur",      16,  32.064,    0.944,    2.60, "yellow"},
  {"Cl", "chlorine",    17,  35.453,    0.994,    2.60, "green"},
  {"Ar", "argon",       18,  39.948,    0.500,    2.60, "aquamarine"},
  {"K",  "potassium",   19,  39.0983,   2.272,    2.60, "pink"},
  {"Ca", "calcium",     20,  40.08,     1.950,    2.60, "snow"},
  {"Sc", "scandium",    21,  44.9559,   1.950,    2.60, "gray"},
  {"Ti", "titanium",    22,  47.90,     1.435,    2.60, "gray"},
  {"V",  "vanadium",    23,  50.9415,   1.311,    2.60, "gray"},
  {"Cr", "chromium",    24,  51.996,    1.250,    2.60, "silver"},
  {"Mn", "manganese",   25,  54.9380,   1.350,    2.60, "gray"},
  {"Fe", "iron",        26,  55.847,    1.250,    2.60, "yellow"},
  {"Co", "cobalt",      27,  58.9332,   1.253,    2.60, "cyan"},
  {"Ni", "nickel",      28,  58.71,     1.246,    2.60, "gray"},
  {"Cu", "copper",      29,  63.54,     1.278,    2.55, "copper"},
  {"Ag", "silver",      47,  107.868,   1.445,    2.87, "silver"},
  {"Au", "gold",        79,  196.967,   1.442,    3.70, "gold"},
  {NULL, NULL,           0,   0.0,      0.000,    0.0,  NULL}
};

/* Look up an entry in the periodic table 
 */
Element *FindElement(char *s)
{
	int i, j;
	if(s != NULL)
	{
	  for(i=0; table[i].name != NULL; ++i) {
	    for(j=0; tolower(s[j]) == tolower(table[i].name[j]); ++j)
	      if(s[j]=='\0')return table + i;
	  }
	}
	return NULL;
}

/* Look up an entry in the periodic table, matching substrings
 * if an exact match isn't found.
 */
Element *GuessElement(char *s)
{
	int i, j;
	Element *e;
	char *sl, el[3];

	if (s == NULL) return table;

	/* Try an exact match first.
	 */
	if((e = FindElement(s)) != NULL) return e;

	/* Try a case-sensitive substring match next.
	 * This gets things like "Si3" or "XSi".
	 */
	for(i=0; table[i].name != NULL; ++i)
	  if(strstr(s, table[i].name) != NULL) return table + i;

	/* Try substring matching a lower-case version,
	 * to handle "si3", etc.
	 */
	for(i=0; table[i].name != NULL; ++i)
	{
	  strncpy(el, table[i].name, 3);
	  for(j=0; el[j] != '\0'; ++j) el[j] = tolower(el[j]);
	  if(strstr(s, el) != NULL) return table + i;
	}

	/* Finally, try a case-insensitive substring match.
	 * We get "SI3" here.
	 */
	for(j=0; s[j] != '\0'; ++j)
	  if(isupper(s[j]))
	  {
	    sl = mystrdup(s);
	    for(j=0; sl[j] != '\0'; ++j) sl[j] = tolower(sl[j]);
	    e = GuessElement(sl);
	    myfree(sl);
	    return e;
	  }

	return table;
}
