/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: fit_cg.c,v 1.2 1996/05/29 21:11:16 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "memory.h"
#include "tools.h"
#include "fit.h"

struct resources {
	Fit fit;
	double ftol;		/* DFP fractional tolerance */
	int iter;		/* DFP iterations */
	double dtol;		/* numerical derivative tolerance */

	int oned_par;		/* global variables for oned_fun() */
	int oned_r;
};

static double oned_fun(double x, void *res)
{
	struct resources *r = res;
	r->fit->par[r->oned_par] = x;
	r->fit->changed = 1;
	r->oned_r = FitEvaluate(r->fit);
	return r->fit->rms;
}

static int cg_fun(double *x, double *f, double *e, int n, void *res)
{
	struct resources *R = res;
	Fit fit = R->fit;
	int i, j, r;
	double savep, e0;

	for(i=j=0; i<fit->npar; ++i)
	  if(fit->fitpar[i])
	  {
	    fit->par[i] = x[j++];
	  }
	fit->changed = 1;

	if(j != n) {
	  dmd_errno = DMD_INVALID_ARGUMENT;
	  return -1;
	}

	if((r = FitEvaluate(fit)) != 0) return r;
	e0 = fit->rms;

	/* Evaluate numerical derivatives.
	 */
	for(i=j=0; i<fit->npar; ++i)
	{
	  if(fit->fitpar[i])
	  {
	    savep = fit->par[i];
	    R->oned_par = i;
	    if(NumericalDerivative(oned_fun, res, savep, e0, R->dtol, f + j))
	    {
	      fprintf(stderr, "FitByConjugateGradient: derivative failed"
		" on parameter %d (%s)\n", i, fit->parname[i]);
	    }
	    f[j] = -f[j];
	    j++;
	    fit->par[i] = savep;
	    fit->changed = 1;
	    if((r = R->oned_r)) return r;
	  }
	}
	*e = e0;

#if 0
	printf("E=%g der=", e0);
	for(i=j=0; i<fit->npar; ++i)
	  if(fit->fitpar[i]) printf(" %g", f[j++]);
	printf("\n");
#endif

	return r;
}

int FitByConjugateGradient(Fit fit, char *options, FILE *fout)
{
	int i, r = 0, np;
	double *x, *f, eps;

	struct resources res;

	res.fit = fit;
	res.ftol = 1.e-5;
	res.iter = 64;
	res.dtol = 1.e-4;

	x = dmalloc(fit->npar);
	for(i=np=0; i<fit->npar; ++i) {
	  if(fit->fitpar[i]) x[np++] = fit->par[i];
	}

	if(np > 0) {
	  f = dmalloc(np);

	  r = optdfp(x, f, &eps, np,
	    1.e10, res.ftol, res.iter, res.iter/4 + 1,
	    NULL, &res, cg_fun, fout);
	  if(r && fout)
	    fprintf(fout, "FitByConjugateGradient: DFP error %d\n", r);

	  for(i=np=0; i<fit->npar; ++i)
	    if(fit->fitpar[i]) fit->par[i] = x[np++];
	  fit->changed = 1;

	  myfree(f);
	}
	myfree(x);

	return r;
}

