/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: fit_gs.c,v 1.2 1996/05/29 21:11:16 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "memory.h"
#include "tools.h"
#include "random.h"
#include "fit.h"

#define VERY_BIG 1.e30

struct resources {
	int ngrid;		/* number of grid points in each direction */
	int maxiter;		/* number of iterations */
};

/* Increment the integers in ic[], so as to run through all of the 
 * combinations with an individual element having a value from 0 to
 * res.ngrid - 1.  Return the sum of the integers.
 */
static int next_ic(int *ic, int np, struct resources *r)
{
	int i = 0, sum = 0;
	while(i < np && ++ic[i] >= r->ngrid) ic[i++] = 0;

	for(i=0; i<np; ++i) sum += ic[i];

	return sum;
}

/* Grid search fit algorithm.
 */
int FitByGridSearch(Fit fit, char *options, FILE *fout)
{
	int i, np, r = 0;
	int iter;
	int *which;
	double *pmin, *dp, f, fbest;
	int *ic, *icbest;

	struct resources res;

	res.ngrid = 8;
	res.maxiter = 6;

	which = imalloc(fit->npar);
	for(i=np=0; i<fit->npar; ++i) if(fit->fitpar[i]) which[np++] = i;
	if(np < 1) {myfree(which); return 0;}

	pmin = dmalloc(np);
	dp = dmalloc(np);
	ic = imalloc(np);
	icbest = imalloc(np);
	
	for(i=0; i<np; ++i)
	  dp[i] = FitTrialRange(fit, which[i]);

	for(iter = 0; iter < res.maxiter; ++iter)
	{
	  for(i=0; i<np; ++i)
	  {
	    pmin[i] = fit->par[which[i]] - 0.5 * dp[i];
	    fit->par[which[i]] = pmin[i];
	    ic[i] = icbest[i] = 0;
	  }
	  fit->changed = 1;
	  if(FitEvaluate(fit))
	    fbest = VERY_BIG;
	  else
	    fbest = fit->rms;

	  /* Evaluate the function at every point within the cube of 
	   * p = [pmin, pmin + dp], and keep track of the best point.
	   */
	  while(next_ic(ic, np, &res))
	  {
	    for(i=0; i<np; ++i)
	      fit->par[which[i]] = pmin[i] + ic[i] * dp[i] / (res.ngrid - 1);
	    fit->changed = 1;

	    if(FitEvaluate(fit) == 0 && (f = fit->rms) < fbest)
	    {
	      fbest = f;
	      for(i=0; i<np; ++i) icbest[i] = ic[i];
	    }
	  }

	  /* Set the parameters at the best point and narrow the range of
	   * the hypercube.
	   */
	  for(i=0; i<np; ++i)
	  {
	    fit->par[which[i]] = pmin[i] + icbest[i] * dp[i] / (res.ngrid - 1);
	    dp[i] /= res.ngrid - 1;
	  }
	  fit->changed = 1;

	  if(fout)
	    fprintf(fout, "FitByGridSearch: "
	      "iteration %d best = %g\n", iter + 1, fbest);
	}

	myfree(icbest);
	myfree(ic);
	myfree(pmin);
	myfree(dp);
	myfree(which);

	return r;
}
