/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: gconnect.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

#include "geometry.h"
#include "memory.h"

#undef DEBUG

#define NOGROUP -1

/* Classifiy all atoms into groups of connected pieces.
 */
static int getgroup(Geometry *g, int *group, double dmax)
{
  int i, j, n, ngroup;
  group[ngroup = 0] = 0; /* define one group, containing the first atom... */
  for(i=1; i<(g->n); ++i)
    group[i] = NOGROUP;  /* ...the rest are unclassified. */

  while(1)
  {
    /* Collect all members of currently defined groups.
     */
    do {
      n = 0; /* number of assignments made */
      for(i=0; i<(g->n); ++i)
	if(group[i] != NOGROUP)
	  for(j=0; j<(g->n); ++j)
	    if(group[j] == NOGROUP && Gdist(g, i, j) < dmax)
	    {
	      group[j] = group[i];
	      ++n;
	    }
    } while(n > 0);

    /* Add a new group if we need to.
     */
    n = 0; /* new groups created (0 or 1) */

    for(i=0; i<(g->n) && group[i] != NOGROUP; ++i);
    if(i < g->n)
      group[i] = ++ngroup;
    else
      break;
  }
  return ngroup + 1;
}

/* Make sure the geometry is connected, in the sense that there are
 * bonds connecting all parts of the molecule.  Move atoms in groups
 * of bonded atoms to ensure that this is the case.
 *
 * David Deaven <deaven@ishmael.ameslab.gov>
 */
Geometry *Gconnect(Geometry *g, double dmax)
{
  int i, j, n, m;
  double d, dmin, dwhich, dx[3], trialdx[3], whichdx[3];
  int *group, ngroup, which;

  group = imalloc(g->n);

  /* Move the closest group to the first group (group 0) as a whole 
   * to connect it to the first group, repeat, until done.
   */
  while((ngroup = getgroup(g, group, dmax)) > 1)
  {
    dwhich = 0.0;
    which = NOGROUP;
    for(n=1; n<ngroup; ++n)
    {
      /* Find the minimum distance and translation vector to move
       * group n into group 0.
       */
      dmin = -1.0;
      for(j=0; j<(g->n); ++j)
	if(group[j] == 0)
	  for(i=0; i<(g->n); ++i)
	    if(group[i] == n)
	    {
		d = Gdisp(g, i, j, trialdx);
		if(d < dmin || dmin < 0.0)
		{
		  dmin = d;
		  for(m=0; m<3; ++m) dx[m] = trialdx[m];
		}
	    }
      
      /* Is this group the best one to move?
       */
      if (dmin < dwhich || which == NOGROUP)
      {
	dwhich = dmin;
	for(m=0; m<3; ++m) whichdx[m] = dx[m];
	which = n;
      }
    }

#ifdef DEBUG
    printf("%d (move %d@%f):", ngroup, which, dmin);
    for(j=0; j<(g->n); ++j) if(group[j] == which)printf(" %d", j+1);
    printf("\n");
#endif

    /* Now whichdx[] is the box-space translation which would carry
     * group "which" atoms into group 0.  But we need only
     * translate a fraction of that to attain dmax.
     */
    for(i=0; i<(g->n); ++i)
      if(group[i] == which)
	for(m=0; m<3; ++m)
	  (g->x + 3*i)[m] += whichdx[m] * 1.01 * (dwhich - dmax) / dwhich;
      
  }

  myfree(group);
  return g;
}
