/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: gdiff.c,v 1.4 1996/05/07 19:36:31 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "geometry.h"
#include "memory.h"
#include "tools.h"

struct gentra_data {
	double xcam[3];
	double vcam[3];
	double scam[9];
	double up[3];
	double t[4*4];
};

/* This is a local part of the next function, gdiff().
 * Given the transformation specified by the static variables above,
 * transform g2 coordinates into cart, and set used[i] equal to
 * the atom number in g1 that corresponds to g2 atom i.
 */
static double gdiff_x(Geometry *g1, double c1[3], Geometry *g2,
	double cart[], int used[], struct gentra_data *gd)
{
  int i, j, n, jmin = 0;
  double rms = 0.0, d, dmin, a;

  /* Generate transformation to be applied to g2 from
   * (xcam, vcam, up), which are assumed valid.  Transform
   * g2's box coordinates into cart[].
   */
  Gentra(gd->xcam, gd->vcam, gd->up, gd->t, gd->scam);
  for(i=0; i<(g2->n); ++i) {
    Transf(g2->x+3*i, gd->t, cart+3*i);
    used[i] = -1;
  }

  /* For each atom in g1, find nearest atom in g2 and use it
   * to compute the contribution to the RMS distance.
   */
  for(i=0; i<(g1->n); ++i) {
    dmin=-1.0;
    for(j=0; j<(g2->n); ++j) {
      if(used[j] < 0) {
	d = 0.0;
	for(n=0; n<3; ++n) {
	  a = (g1->x+3*i)[n] - c1[n] - (cart+3*j)[n];
	  d += a*a;
	}
	if(dmin < 0.0 || d < dmin) {
	  dmin = d;
	  jmin = j;
	}
      }
    }
    if(dmin < 0.0)
    {
      fprintf(stderr, "gdiff_x: unexpected condition: dmin=%g\n", dmin);
      return -1.0;
    }
    used[jmin] = i;
    rms += dmin;
  }
  rms = sqrt(rms/((double) g1->n));
  return rms;
}

/* Given two geometries,
 * 1. Translate the second to the first's center of number.
 * 2. If possible, rotate the second to minimize the RMS atomic
 *    distances between the atoms in correspondence.
 * 3. Return the RMS distance, and leave the second geometry
 *    transformed to the position, with re-indexed atoms.
 *
 * The linear resolution determines the angular resolution through
 * the radius of the geometry.
 */
#define NSEARCH 8 /* angular divisions in search */
#define RESOLUTION 0.05 /* linear resolution */
double Gdiff(Geometry *g1, Geometry *g2) 
{
  return Gdiff_spec(g1, g2, NSEARCH, RESOLUTION);
}

double Gdiff_spec(Geometry *g1, Geometry *g2, int ngrid, double res)
{
  int i, j, m, imin = 0, jmin = 0;
  int *used;
  double *cart;
  double theta, xi, rms, rmsmin = 0.0;
  double theta0, theta1, dtheta, xi0, xi1, dxi=0.0;
  double rmax;
  struct gentra_data *gd;
  static double c1[3], c2[3];

  if(g1->n != g2->n) return 1.e10; /* pretty dang different! */

  /* We need some scratch space */
  used = imalloc(g2->n);
  cart = dmalloc(3 * g2->n);
  gd = mymalloc(sizeof(*gd));

  /* Find the center of number for the geometries.
   */
  gd->xcam[0] = gd->xcam[1] = gd->xcam[2] = 0.0;
  c1[0] = c1[1] = c1[2] = 0.0;
  c2[0] = c2[1] = c2[2] = 0.0;
  for(i=0; i<(g1->n); ++i)
  {
    for(m=0; m<3; ++m)
    {
      c1[m] += (g1->x+i*3)[m]; 
      c2[m] += (g2->x+i*3)[m];
    }
  }
  for(m=0; m<3; ++m) 
  {
    c1[m] /= (double) g1->n;
    c2[m] /= (double) g1->n;
    gd->xcam[m] = c2[m];
  }
  /* Find the maximum radius of g2 to determine resolution.
   */
  rmax = 0.0;
  for(i=0; i<(g1->n); ++i)
  {
    rms = 0.0;
    for(m=0; m<3; ++m)
      rms += ((g2->x+3*i)[m] - c2[m]) * ((g2->x+3*i)[m] - c2[m]);
    if(rms > rmax) rmax = rms;
  }
  if(rmax <= 0.0)
  {
    fprintf(stderr, "gdiff: rmax=%g -- pathological case\n", rmax);
    Gprint(stdout, g2);
    rmax = 1.0;
  }
  rmax = res / sqrt(rmax); /* convert to angular measure */

  /* Start with the entire range, then narrow the search
   * until we get the desired resolution.
   * If the geometries are periodic in more than one
   * direction, or are incomplete, we cannot rotate them.
   */
  theta0 = xi0 = 0.0;
  if((g1->period & (GP_PERIODIC | GP_INCOMPLETE)) || 
     (g2->period & (GP_PERIODIC | GP_INCOMPLETE))    )
  {
	theta1 = theta0;
	xi1 = xi0;
  } else {
	theta1 = M_PI;
	xi1 = 2.0*M_PI;
  }

  while((dxi = xi1 - xi0, dtheta = theta1 - theta0) > rmax)
  {
    dtheta /= (double) ngrid;
    dxi /= (double) ngrid;

    /* Explore all rotations, determine minimum rotation
     * direction.  This is a double search loop.
     */
    for(i=0; i<=ngrid; ++i)
    {
      theta = theta0 + ((double) i)*dtheta;
      gd->vcam[0] = 0.0; gd->vcam[1] = sin(theta); gd->vcam[2] = -cos(theta);
      for(j=0; j<ngrid; ++j)
      {
	xi = xi0 + ((double) j)*dxi;
	gd->up[0] = sin(xi); gd->up[1] = cos(xi); gd->up[2] = 0.0;
	if((rms = gdiff_x(g1, c1, g2, cart, used, gd)) < 0.0)
	{ 
	  myfree(used);
	  myfree(cart);
	  myfree(gd);
	  return 1.e10;
        }
	if((i==0 && j==0) || rms < rmsmin)
	{
	  rmsmin = rms;
	  imin = i;
	  jmin = j;
	}
      }
    }
    if(imin > 0) theta0 += ((double) imin-1)*dtheta;
    theta1 = theta0 + 2.0*dtheta;
    if(jmin > 0)xi0 += ((double) jmin-1)*dxi;
    xi1 = xi0 + 2.0*dxi;
  }

  /* Return the geometry rotated to the minimal RMS direction,
   * and re-ordered so as to match up by number.
   */
  theta = theta0;
  gd->vcam[0] = 0.0; gd->vcam[1] = sin(theta); gd->vcam[2] = -cos(theta);
  xi = xi0;
  gd->up[0] = sin(xi); gd->up[1] = cos(xi); gd->up[2] = 0.0;
  rms = gdiff_x(g1, c1, g2, cart, used, gd);

  for(i=0; i<(3 * g2->n); ++i) g2->x[i] = cart[i] + c1[i % 3];
  (void) Greorder(g2, used);

  myfree(used);
  myfree(cart);
  myfree(gd);

  return rms;
}
