/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: geometry.h,v 1.19 1996/07/09 16:36:10 deaven Exp deaven $
 * $Author: deaven $
 */

#ifndef _GEOMETRY_H_
#define _GEOMETRY_H_

#include <stdio.h>
#include "dmd/tools.h"

#ifdef __cplusplus
extern "C" {
#endif

/* periodicity flags */
#define GP_PDIR(M)	(1 << M)
#define GP_PX		GP_PDIR(0)
#define GP_PY		GP_PDIR(1)
#define GP_PZ		GP_PDIR(2)
#define GP_PERIODIC	(GP_PX | GP_PY | GP_PZ)
#define GP_INCOMPLETE	(1 << 3) /* this geometry is only a part of the whole */

/* All information about an atomic geometry.  The coordinates are
 * stored in box coordinates.  Physical coordinates are obtained
 * by left-multiplication of the box[], and box coordinates can be
 * obtained by left-multiplication of the inverse box: boxi[].
 *
 * The atomic types are stored as an array of pointers to character
 * strings, usually denoting the atomic species.
 *
 * The flag bits in "period" (defined above in GP_*) control whether
 * the box is periodically repeated in each of the three directions,
 * or also if this geometry is "incomplete". This latter bit is used
 * when only a subset of a complete system is represented with a 
 * Geometry structure.  In this case, any operation which might be
 * performed on a complete geometry specification (e.g. rotation)
 * should not be performed.
 *
 * The integer flags[] can either be unallocated (NULL), or allocated
 * and the meaning controlled by the application.  In any case, if
 * they are allocated, the library routines will preserve them and
 * any operation which re-orders the atoms will also re-order the 
 * flags.
 */
typedef struct {
	int n;			/* number of atoms */
	double *x;		/* box coordinates */
	double *box;		/* conversion to physical units */
	double *boxi;		/* inverse of box */
	char *name;		/* name of the geometry as a whole */
	char **aname;		/* name of each atom */
	int period;		/* periodicity flags */
	int *flags;		/* optional array of flags */
} Geometry;

/* Atoms in a geometry can be selected using a selection mask.
 */
typedef struct {
	int n;			/* number of atoms */
	unsigned char *s;	/* array of bits */
} GeometryMask;
#define GMaskIsSelected(MASK,ATOM) \
  ((MASK) != NULL && (MASK)->n > (ATOM) && (MASK)->s[ATOM])

/* Function prototypes.
 */
void Gfree(Geometry *g);
void GprintXYZ(FILE *f, Geometry *g);
void Gprintbox(FILE *f, Geometry *g);
void Gprint(FILE *f, Geometry *g);
Geometry *Greorder(Geometry *g, int *order);
void Gcopyatom(Geometry *dest, int i, Geometry *src, int j);
Geometry *Gstdorient(Geometry *g);
int Gmoment(Geometry *g, double *mass,
	double mom[3], double x0[3], double axes[3*3]);
Geometry *Gdup(Geometry *src);
Geometry *Gcpy(Geometry *dst, Geometry *src);
Geometry *Gcat(Geometry *dst, Geometry *g1, Geometry *g2);
Geometry *Gmalloc(int n, char *name);
Geometry *Grealloc(Geometry *old, int n);
Geometry *Gread(FILE *f);
Geometry *GreadWithError(FILE *f, FILE *ferr);
double Gdist(Geometry *g, int i, int j);
double Gdisp(Geometry *g, int i, int j, double dx[3]);
double Gdisp2(Geometry *g, int i, int j, double dx[3]);

double Gdiff(Geometry *g1, Geometry *g2);
double Gdiff_spec(Geometry *g1, Geometry *g2, int ngrid, double res);
double GAlignYPR(Geometry *g1, Geometry *g2,
	double ypr[3], double dx[3], int ngrid, double res);
Geometry *Glattice(Geometry *g, 
  int nx0, int nx1, int ny0, int ny1, int nz0, int nz1);
Geometry *Grrotate(Geometry *g);
Geometry *GboxCenter(Geometry *g);
double *GgetBoxCenter(Geometry *g, double center[3]);
Geometry *Gcontract(Geometry *g, double dmax);
Geometry *Gconnect(Geometry *g, double dmax);
Geometry *Gseparate(Geometry *g, double dmin);

unsigned char *Gdigitize(Geometry *g, unsigned long *size);
Geometry *Gundigitize(unsigned char *buffer, unsigned long size);
unsigned char *Gdigitize_ansi(Geometry *g, unsigned long *size);
unsigned char *Gdigitize_buf_ansi(Geometry *g, unsigned char *buffer,
	unsigned long *offset, unsigned long *size);
Geometry *Gundigitize_ansi(unsigned char *buffer, unsigned long size);
Geometry *Gundigitize_buf_ansi(unsigned char *buffer,
	unsigned long *offset, unsigned long size);

Geometry *Gsetbox(Geometry *g, double *newbox);
Geometry *GsetboxStandard(Geometry *g);
Geometry *GsetboxRect(Geometry *g);
Geometry *GsetboxCube(Geometry *g);
double *Genclosebox(Geometry *g, double box[3*3]);
double Gvolume(Geometry *g);

struct Gfile *GFileOpen(char *path);
void GFileInfo(struct Gfile *gf, FILE **f, char **path,
	int *frame, int *maxframe);
int GFileAdvance(int n, struct Gfile *gf);
int GFileRewind(int n, struct Gfile *gf);
Geometry *GFileRead(struct Gfile *gf);
int GFileReopen(struct Gfile *gf);
void GFileClose(struct Gfile *gf);

Geometry *GreadMPCOutput(FILE *f);
void GprintMPC(FILE *f, Geometry *g);

GeometryMask *GMask(int n);
void GMaskDestroy(GeometryMask *mask);
GeometryMask *GMaskOr(GeometryMask *a, GeometryMask *b);
GeometryMask *GMaskAnd(GeometryMask *a, GeometryMask *b);
GeometryMask *GMaskRead(FILE *f);
void GMaskPrint(FILE *f, GeometryMask *mask);
GeometryMask *GMaskResize(GeometryMask *old, int n);
GeometryMask *GMaskCopy(GeometryMask *dst, GeometryMask *src);

double *GSetYPRFromFlags(Geometry *g, double *ypr);
Geometry *GSetFlagsFromYPR(Geometry *g, double *ypr);

Geometry *GCentralCell(Geometry *g);

Geometry *GPrettyCell(Geometry *g, double rbond);

Geometry **GSplit(Geometry *g,
	int (*ismember)(Geometry *group, Geometry *test, int atom));

#ifdef __cplusplus
}
#endif

#endif /* _GEOMETRY_H_ */
