/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: gmask.c,v 1.4 1995/12/19 18:09:23 deaven Exp $
 * $Author: deaven $
 */

#include "geometry.h"
#include "memory.h"

/* Create a new mask which can hold n entries.
 */
GeometryMask *GMask(int n)
{
	int i;
	GeometryMask *new = mymalloc(sizeof(*new));
	new->n = n;
	new->s = mymalloc(n * sizeof(*(new->s)));
	for(i=0; i<n; ++i) new->s[i] = 0;
	return new;
}

/* Deallocate a selection mask.
 */
void GMaskDestroy(GeometryMask *mask)
{
	if(mask) {
	  myfree(mask->s);
	  myfree(mask);
	}
}

/* Resize a selection mask.
 */
GeometryMask *GMaskResize(GeometryMask *old, int n)
{
	if(old == NULL) old = GMask(n);
	if(old->n != n) {
	  old->s = myrealloc(old->s, n * sizeof(*(old->s)));
	  old->n = n;
	}
	return old;
}

/* Copy the selection mask src to dst, resizing dst if required.
 */
GeometryMask *GMaskCopy(GeometryMask *dst, GeometryMask *src)
{
	int i;
	dst = GMaskResize(dst, src->n);
	for(i=0; i<src->n; ++i) dst->s[i] = src->s[i];
	return dst;
}

/* Perform a logical AND operation on two masks, returning the
 * results in a, which will be resized if b is bigger, or
 * allocated is a is NULL.
 */
GeometryMask *GMaskAnd(GeometryMask *a, GeometryMask *b)
{
	int i;
	if(a == NULL || a->n < b->n) a = GMaskResize(a, b->n);
	for(i=0; i<b->n; ++i) a->s[i] &= b->s[i];
	for(i=b->n; i<a->n; ++i) a->s[i] = 0;
	return a;
}

/* Perform a logical OR operation on two masks, returning the
 * results in a, which may be resized if b is bigger, or
 * allocated is a is NULL.
 */
GeometryMask *GMaskOr(GeometryMask *a, GeometryMask *b)
{
	int i;
	if(a == NULL || a->n < b->n) a = GMaskResize(a, b->n);
	for(i=0; i<b->n; ++i) a->s[i] |= b->s[i];
	return a;
}

/* Print the geometry selection mask to f, in such a way that
 * GMaskRead() can read it in.  Only the selected numbers
 * are written.
 */
void GMaskPrint(FILE *f, GeometryMask *mask)
{
	int i, nsel;
	for(i=nsel=0; i<mask->n; ++i) if(mask->s[i]) ++nsel;
	fprintf(f, "#%d\n", nsel);
	for(i=nsel=0; i<mask->n; ++i)
	  if(mask->s[i]) 
	    fprintf(f, "%d%s", i, ++nsel % 8 ? " " : "\n");
	if(nsel % 8) fprintf(f, "\n");
}

/* Read a geometry selection mask from f.
 */
GeometryMask *GMaskRead(FILE *f)
{
	int i, n, nsel, *on;
	GeometryMask *new;

	if(fscanf(f, "#%d", &nsel) != 1) return NULL;

	on = imalloc(nsel);

	for(i=n=0; i<nsel; ++i) {
	  if(fscanf(f, "%d", on + i) != 1) break;
	  if(on[i] < 0) break;
	  if(on[i] > n) n = on[i];
	}
	if(i != nsel) {
	  myfree(on);
	  return NULL;
	}

	new = GMask(n + 1);
	for(i=0; i<nsel; ++i) new->s[on[i]] = 1;
	myfree(on);
	return new;
}
