/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: gmoment.c,v 1.5 1996/04/18 22:25:32 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "geometry.h"
#include "tools.h"
#include "eisdiag.h"
#include "memory.h"

/* Compute the center of mass and moment of inertia for the geometry.
 * if mass == NULL all masses are assumed equal to one.
 * mom[i] = ith moment, x0 = c.o.m., and
 * axes[i][j] = jth component of principal axis i.
 * The moments are returned in descending order.
 *
 * Return value is zero, or non-zero if an error occured.
 *
 * David Deaven ishmael.ameslab.gov 
 */
int Gmoment(Geometry *g, double *mass,
	double mom[3], double x0[3], double axes[3*3])
{
	int i, j, l, m, n;
	static double x[3], a[3*3], mt;

	/* find center of mass
	 */
	mt = x0[0] = x0[1] = x0[2] = 0.0;
	for(i=0; i<(g->n); ++i) mt += (mass == NULL ? 1.0 : mass[i]);
	for(j=0; j<(g->n); ++j) {
	  mmul3(g->box, g->x+j*3, x);
	  for(i=0; i<3; ++i) x0[i] += x[i] * (mass == NULL ? 1.0 : mass[j]);
	} 
	for(i=0; i<3; ++i) x0[i] /= mt;

	/* set up mass tensor 
	 */
	for(i=0; i<9; ++i) a[i] = 0.0;

	for(l=0; l<(g->n); ++l)
	{
	  mt = (mass == NULL ? 1.0 : mass[l]);
	  mmul3(g->box, g->x+l*3, x);
	  for(i=0; i<3; ++i) {
	    for(j=0; j<3; ++j)
	    {
	      a[i*3 + j] -= mt * (x[i]-x0[i]) * (x[j]-x0[j]);
	      if(i==j) a[i*3 + j] += mt * (
		(x[0]-x0[0]) * (x[0]-x0[0]) +
		(x[1]-x0[1]) * (x[1]-x0[1]) +
		(x[2]-x0[2]) * (x[2]-x0[2])  );
	    }
	  }
	}

	/* Diagonalize.
	 */
	if(eistred2(a, 3, mom, x)) return -1;
	if(eistql2(mom, x, 3, a)) return -1;

	/* a[i][j] is now ith component of eigenvector j
	 * Sort axes by moments (in a kludgey way XXX).
	 */
	for(i=l=0; i<3; ++i)
	  if(mom[i] > mom[l]) l=i;
	m=0; if(m==l)++m;
	for(i=m+1; i<3; ++i)if(i!=l && mom[i] > mom[m])m=i;
	n=0; while(n==l || n==m)++n;
	for(i=n+1; i<3; ++i)if(i!=l && i!=m && mom[i] > mom[n])n=i;

	x[0] = mom[l]; x[1] = mom[m]; x[2] = mom[n];
	for(i=0; i<3; ++i) {
	  axes[i  ] = a[l*3 + i];
	  axes[i+3] = a[m*3 + i];
	  axes[i+6] = a[n*3 + i];
	  mom[i] = x[i];
	}

	return 0;
}

