/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: histogram.c,v 1.3 1996/02/23 00:42:31 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <stdlib.h>
#include "histogram.h"
#include "memory.h"
#include "tools.h"

/* Allocate a histogram.
 */
Histogram *Histmalloc(double dx)
{
	Histogram *h = mymalloc(sizeof(Histogram));
	h->n = 0;
	h->dx = dx;
	h->base = 0.0;
	h->data = NULL;
	return h;
}

/* Clear a histogram
 */
Histogram *Histclear(Histogram *h)
{
	myfree(h->data);
	h->data = NULL;
	h->n = 0;
	return h;
}

/* Free the memory used by a histogram
 */
void Histfree(Histogram *h)
{
	Histclear(h);
	myfree(h);
}

/* Add a delta function to a histogram.
 * Return the index to the bin number that was accessed,
 * which is only valid until the next call into this code.
 */
int Histadd(Histogram *h, double x, double z)
{
	int i, m;
	m = anint((x - h->base) / h->dx);
	for(i=0; i<h->n; ++i)
	{
		if(m == h->data[i].m)
		{
			h->data[i].z += z;
			return i;
		}
	}
	h->data = myrealloc(h->data, (h->n + 1) * sizeof(Bindata));
	h->data[h->n].m = m;
	h->data[h->n].z = z;

	return (h->n)++;
}

static int cmp(const void *a, const void *b)
{
	int ma = ((Bindata *) a)->m;
	int mb = ((Bindata *) b)->m;
	if (ma < mb) return -1;
	if (ma > mb) return 1;
	return 0;
}

/* Sort a histogram into ordered ordinates.
 */
void Histsort(Histogram *h)
{
	unsigned int n = h->n;
	if(h->n > 0)
	  qsort(h->data, n, sizeof(Bindata), cmp);
}

/* Produce xyplot() instructions to make a histogram plot.
 */
void HistXYoff(FILE *f, Histogram *h, double xoff, double yoff)
{
	int i;
	double x, dx = h->dx;

	Histsort(h);

	fprintf(f, "/* histogram xyplot data */\n");
	for(i=0; i<h->n; ++i)
	{
	  x = ((double) h->data[i].m) * dx + h->base;
	  if(i==0 || h->data[i].m > 1 + h->data[i-1].m)
	    fprintf(f, "%g %g ", x + xoff - 0.5*dx, yoff);
	  fprintf(f, "%g %g %g %g",
	    x + xoff - 0.5*dx, h->data[i].z + yoff,
	    x + xoff + 0.5*dx, h->data[i].z + yoff);
	  if(i==h->n - 1 || h->data[i+1].m > 1 + h->data[i].m)
	    fprintf(f, " %g %g", x + xoff + 0.5*dx, yoff);
	  fprintf(f, "\n");
	}
	fprintf(f, "/* end of histogram data */\n");
}

void HistXY(FILE *f, Histogram *h)
{
	HistXYoff(f, h, 0.0, 0.0);
}
