/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: linmin.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "memory.h"
#include "tools.h"

/* This structure is used to provide a local copy of the one
 * dimensional line minimization during the calls to lower-level
 * routines which think that they are operating on functions
 * of a single variable.  By allocating a distinct `environment'
 * structure for each call, we can support multiply-threaded use
 * of this code.
 */
struct fun_env {
	int refresh;	/* flag to force calculation */
	int n;		/* number of parameters */
	double *p;	/* parameter values */
	double *dp;	/* derivative values */
	double *xi;	/* direction to move along */
	double xc;	/* current position */
	double fc;	/* current function value */
	int (*fun)(double *, double *, double *, int);
};

/* Driver function.  Given an `environment,' compute the 
 * function at position x along the direction xi in parameter
 * space.  Remember where we are, and take care not to
 * evaluate the function at the same point twice in a
 * row (i.e., the caller can get the function and the 
 * derivative as many times as required with no extra work).
 */
static int fun_fun(double x, double *y, double *dy, void *fev)
{
  struct fun_env *fe = fev;
  int i, r = 0;
  if(x != fe->xc || fe->refresh)
  {
    for(i=0; i<fe->n; ++i)
      fe->p[i] += (x - fe->xc) * fe->xi[i];
    r = (*fe->fun)(fe->p, fe->dp, &(fe->fc), fe->n);
    fe->xc = x;
    fe->refresh = 0;
  }
  *y = fe->fc;
  for(*dy=0.0, i=0; i<fe->n; ++i)
    *dy -= fe->xi[i] * fe->dp[i];

  return r;
}

static double *allocate_copy(double *p, int n)
{
	int i;
	double *new = dmalloc(n);
	for(i=0; i<n; ++i) new[i] = p[i];
	return new;
}

/* Minimize a passed function along a line.  Return the point that
 * minimizes the function, the value of the function at that point,
 * and the derivative at that point
 *
 * Input:
 *   p[] = parameters.
 *   xi[] = direction to minimize along in parameter space.
 *   n = number of parameters
 *   tol = fractional tolerance in value
 *   iter = maximum iterations.
 *   fun() = function to compute function and derivative.
 *
 * Output:
 *   fret = minimized function.
 *   p[] = new parameters.
 *   f[] = derivative at p[] (new force).
 *   return value is -1 if convergence was not achieved, or
 *     a non-zero code from fun().
 *
 * This routine supports multiple threads of control, finally.
 */
int linmin(double *p, double *xi, int n, double tol, int iter,
  int (*fun)(double *, double *, double *, int), double *fret, double *f)
{
    int i, r;
    double fa, fb, ax, bx, fx, xmin, xx, df;
    struct fun_env *fe;

    /* Set up a fun_fun() environment to minimize along the positive
     * xi[] direction with positive x values.
     */
    fe = mymalloc(sizeof(struct fun_env));
    fe->n = n;
    fe->xi = allocate_copy(xi, n);
    fe->p = allocate_copy(p, n);
    fe->dp = dmalloc(n);
    fe->fun = fun;
    fe->xc = 0.0;
    fe->refresh = 1; /* we must do the evaluation the first time */

    /* Bracket the minimum.  We assume that the current point
     * is valid, and try to find a second point along the 
     * gradient to pass to mnbrak().
     */
    ax = 0.0;
    if((r = fun_fun(ax, &fa, &df, fe))) goto done;
    for(xx = 0.1; xx > 1.e-6; xx *= 0.1)
      if((r = fun_fun(xx, &fx, &df, fe)) == 0) break;
    if(r) goto done;

    if((r = mnbrak(&ax, &xx, &bx, &fa, &fx, &fb, fun_fun, fe)))goto done;
    if((r = fun_fun(xx, &fx, &df, fe))) goto done;

    /* Converge on the minimum.
     */
    r = dbrent(ax, xx, bx, fx, df, fun_fun, fe, tol, iter, &xmin, &fx);
    if(r == -1) r = 0; /* ignore "too many iters" */
    if(r) goto done;

    for(i=0; i<fe->n; ++i)
    {
      p[i] = fe->p[i] + fe->xi[i] * (xmin - fe->xc);
      f[i] = fe->dp[i];
    }
    *fret = fx;

    r = 0;

done:
    myfree(fe->dp);
    myfree(fe->p);
    myfree(fe->xi);
    myfree(fe);
    return r;
}

