/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: lstsquare.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "tools.h"
#include "memory.h"

static void gajo (double **matrix, int mrows, int ncols)
/*
c reduces a matrix to reduced row - echelon form
c
c matrix  - the input matrix (m rows x n columns)
c mrows   - mrows in matrix
c ncols   - columns in matrix
c m,n	  - dimensioned rows,columns
c
*/
{
  int i, j, irow;
  int toprow, leftcol, botrow, rightcol;
  double a, b;

	toprow = leftcol = 0;

/* I. Find leftmost non-zero column */

  do {
	irow = toprow;
	while (matrix[irow][leftcol] == 0.0) {
	  if((++irow) >= mrows) {
	    irow = toprow;
	    if((++leftcol) >= ncols)return;
	  }
	}

/*
 * II. Interchange rows to get non-zero entry in row toprow
 *     Also, insure that toprow is normalized
 */

	a = 1.0/matrix[irow][leftcol];
	for(i=leftcol; i<ncols; ++i) {
	  b = matrix[irow][i]*a;
	  matrix[irow][i] = matrix[toprow][i];
	  matrix[toprow][i] = b;
	}

/*
 * III. Add multiples of top row to below rows to get zero in
 *      column icol
 */

	if (toprow != mrows) {
	  for(i=toprow+1; i<mrows; ++i) {
	    if (matrix[i][leftcol] != 0.0) {
	      a = -matrix[i][leftcol]/matrix[toprow][leftcol];
	      for(j=leftcol; j<ncols; ++j) matrix[i][j] += a*matrix[toprow][j];
	    }
	  }
	}

/*
 * IV. ignore toprow, do next leftmost column
 */
	++leftcol; ++toprow;

  } while(leftcol < ncols && toprow < mrows);

/*
 * V. Matrix is now in row-echelon form. Reduce:
 */

  botrow = mrows-1;

/*
 * VI. Find bottom non-zero row
 */

  do {
	rightcol = 0;
	while (matrix[botrow][rightcol] == 0.0) {
	  if ((++rightcol) >= ncols) {
	    rightcol = 0;
	    if ((--botrow) < 0) return;
	  }
	}

/*
 * VII. Add multiples of this row to rows above to eliminate
 *      non-zero entries in leading ones column
 */

	for(i=botrow-1; i>=0; --i) {
	  a = -matrix[i][rightcol]/matrix[botrow][rightcol];
	  for(j=rightcol; j<ncols; ++j) matrix[i][j] += a*matrix[botrow][j];
	}

/*
 * VIII. Ignore bottom row, do it again for upper submatrix
 */

  } while((--botrow) > 0);

}

/* Fit a set of input x,y points to the best possible (least sqared
 * error) polynomial of order 'n'.
 *
 * x[], y[], nd -- input data and number of supplied points.
 * coef[]       -- coefficient vector, coef[0] to coef[order].
 * order        -- order of approximation desired. If zero, the solution
 *                 will be in the form y=ax, where coef[0]=a.
 * err          -- mean error in approximation.
 * rmserr	-- root mean sqare average error in approximation.
 * s		-- standard deviation for err in approximation.
 *
 * David Deaven, from old code written at the Naval Research Laboratory,
 *               circa 1987.
 */
void lstsquare(double x[], double y[], int nd, 
	double *coef, int order, double *err, double *rmserr, double *s)
{
  int i, j, n, simple=0;
  double *xsum, *xysum, ysum, y2sum, *matrix_s, **matrix;

  if (order==0) simple = n = 1;
  else n = order;

  xsum = mymalloc((2*n+1)*sizeof(double));
  xysum = mymalloc((n+1)*sizeof(double));
  matrix_s = mymalloc((n+1)*(n+2)*sizeof(double));
  matrix = mymalloc((n+1)*sizeof(double*));
  if(xsum==NULL || xysum==NULL || matrix_s==NULL || matrix==NULL) {
    fprintf(stderr, "cannot allocate memory for order %d\n", n);
    exit(1);
  }
  for(i=0; i<=n; ++i)matrix[i] = matrix_s + i*(n+2);

/* Set up/initialize variables */

  y2sum = ysum = 0.0;
  for(i=0; i<=n; ++i) xysum[i] = 0.0;
  for(i=0; i<=(2*n); ++i) xsum[i] = 0.0;

/* calculate summations */

  for(j=0; j<nd; ++j) {
    for(i=0; i<=(2*n); ++i) xsum[i] += pow(x[j], (double) i);
    for(i=0; i<=n; ++i) xysum[i] += y[j]*pow(x[j], (double) i);
    y2sum += y[j]*y[j];
    ysum += y[j];
  }

/* Solve the simple equations if that's what's needed */

  if (simple) {
	  double a = xysum[1]/xsum[2];
	  (*err) = (a*xsum[1]-ysum)/xsum[0];
	  (*rmserr) = sqrt((a*a*xsum[2]-2.0*a*xysum[1]+y2sum)/xsum[0]);
	  (*s) = a*a*xsum[2] + y2sum + xsum[0]*(*err)*(*err);
	  (*s) += 2.0*((*err)*ysum - (*err)*a*xsum[1] - a*xysum[1]);
	  (*s) = sqrt((*s)/xsum[0]);
	  coef[0] = a;
  } else {

/* Otherwise, solve matrix system.
 * Coefficient matrix is symmetric so entry i,j is xsum(i+j) 
 */

  for(i=0; i<=n; ++i) {
    for(j=0; j<=n; ++j) matrix[i][j] = xsum[i+j];
    matrix[i][n+1] = xysum[i];
  }
  gajo(matrix, n+1, n+2);

/* Transfer matrix solutions to coef(0:n) */

  for(i=0; i<=n; ++i)coef[i] = matrix[i][n+1];

/* Calculate error quantities */

	(*err) = -ysum;
	for(i=0; i<=n; ++i)(*err) += coef[i] * xsum[i];
	(*err) /= xsum[0];

	(*rmserr) = 0.0; /* a summation of sqared terms and perm. terms */
	for(i=0; i<n; ++i) {
	  for(j=i+1; j<=n; ++j) (*rmserr) += coef[i]*coef[j]*xsum[i+j];
	}
	for(i=0; i<=n; ++i) (*rmserr) -= coef[i]*xysum[i];
	(*rmserr) *= 2.0; /* permutations are doubled */

	for(i=0; i<=n; ++i) (*rmserr) += xsum[2*i]*(coef[i]*coef[i]);
	(*rmserr) += y2sum;

	(*s) = (*rmserr) + (*err)*(*err) + (*err)*ysum;
	(*rmserr) = sqrt(fabs((*rmserr)/xsum[0]));

	for(i=0; i<=n; ++i) (*s) -= 2.0*coef[i]*xsum[i]*(*err);
	(*s) = sqrt(fabs((*s)/xsum[0]));
  }

  myfree(xsum); myfree(xysum); myfree(matrix_s); myfree(matrix);
}
