/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: marquardt.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "tools.h"
#include "memory.h"

/* Marquardt nonlinear least squares fitting, adapted from 
 * _Numerical Recipes in C_ 
 *
 * David Deaven deaven@ishmael.ameslab.gov
 */

/* Evaluate the linearized fitting matrix in Marquardt()
 * Compute linearized fitting matrix alpha, vector beta.
 * Compute chi^2.
 */
static void mrqcof(double x[], double y[], double sig[], int ndata,
	double a[], int ia[], int ma,
	double alpha[], double beta[], double *chisq,
	void (*fun)(double,double *,double *,double *,int))
{
  int i, jl, j, kl, k, mfit;
  double ymod, wt, sig2i, dy, *dyda;
 
  /* Intialize matrix, vector to zero.
   */
  for (mfit=jl=0; jl<ma; ++jl) if(ia[jl]) beta[mfit++] = 0.0;
  for (j=0; j<(mfit*mfit); ++j) alpha[j] = 0.0;

  /* Sum over all data.
   */
  dyda = dmalloc(ma);
  *chisq = 0.0;
  for (i=0; i<ndata; ++i)
  {
    (*fun)(x[i], a, &ymod, dyda, ma);
    sig2i = 1.0/(sig[i]*sig[i]);
    dy = y[i]-ymod;
    for(j=jl=0; jl<ma; ++jl) if(ia[jl])
    {
      wt = dyda[jl]*sig2i;
      for(k=kl=0; kl<ma; ++kl) if(ia[kl])
      {
	alpha[j*mfit + k] += wt*dyda[kl];
	++k;
      }
      beta[j] += dy*wt;
      ++j;
    }
    *chisq += dy*dy*sig2i;
  }
  myfree(dyda);

  /* Fill in symmetric side.
   */
  for (j=1; j<mfit; ++j)
    for (k=0; k<j; ++k)
      alpha[k*mfit + j] = alpha[j*mfit + k];

}

/* Callable top-level Marquardt optimizer.
 * Perform general nonlinear least-squares fitting.
 * 
 * Input:
 *   x[], y[], sig[] = data and standard deviations, of which there 
 *	are ndata.
 *   a[] = initial parameters, of which there are ma.
 *   ia[] = flags for optimization of parameters. ia[i] == 0 means
 *	do not change the valus of parameter i.  The number of non-zero
 *	entries in ia[], "mfit", gives the minimal dimension of some
 *	of the output arrays.
 *   fun(x, a, y, dyda) = model function returning y and dyda as a
 *	function of x and a.
 *   alamda = negative to initialize, zero to return covar and 
 *	alpha.
 *
 * Output:
 *   covar = mfit*mfit covarience matrix for optimized parameters.
 *   alpha = mfit*mfit curvature matrix for optimized parameters.
 *   chisq = chi^2
 *   alamda = step size.
 *
 */
int Marquardt(double x[], double y[], double sig[], int ndata,
	double a[], int ia[], int ma,
        void (*fun)(double, double *,double *,double *,int),
	double *covar, double *alpha, double *chisq, double *alamda)
{
  int jl, j, k;
  static int mfit;
  static double *da = NULL, *atry = NULL, *beta = NULL, ochisq;

  /* Initialize 
   */
  if (*alamda < 0.0)
  {
     *alamda=0.001;

     for (mfit=jl=0; jl<ma; ++jl) if(ia[jl]) ++mfit;

     atry = dmalloc(ma);
     da = dmalloc(mfit);
     beta = dmalloc(mfit);

     mrqcof(x, y, sig, ndata, a, ia, ma, alpha, beta, chisq, fun);
     ochisq = *chisq;
     for(jl=0; jl<ma; ++jl) atry[jl] = a[jl];
  }

  /* Alter linearized fitting matrix, by augmenting 
   * diaginal elements.
   */
  for(j=jl=0; jl<ma; ++jl) if(ia[jl])
    {
      for(k=0; k<mfit; ++k) covar[j*mfit + k] = alpha[j*mfit + k];
      covar[j*mfit + j] = alpha[j*mfit + j]*(1.0+(*alamda));
      da[j] = beta[j];
      ++j;
    }

  /* Solve matrix
   */
  if((j=gaussj(covar, mfit, da, 1))) return j;

  /* Return when converged.
   */
  if (*alamda == 0.0)
  {
     myfree(beta); 
     myfree(da);
     myfree(atry);
     return 0;
  }

  /* Check to see if the trial succeeded
   */
  for (j=jl=0; jl<ma; ++jl) if(ia[jl])
    atry[jl] = a[jl] + da[j++];

  mrqcof(x, y, sig, ndata, atry, ia, ma, covar, da, chisq, fun);
  if (*chisq < ochisq)
  {
     *alamda *= 0.1;
     ochisq=(*chisq);
     for (j=jl=0; jl<ma; ++jl) if(ia[jl])
     {
        for (k=0; k<mfit; ++k) alpha[j*mfit + k] = covar[j*mfit + k];
        beta[j] = da[j];
        a[jl] = atry[jl];
	++j;
     }
  } else {
     *alamda *= 10.0;
     *chisq = ochisq;
  }
  return 0;
}

#define SWAP(a,b) {double temp=(a);(a)=(b);(b)=temp;}

/* Gauss-Jordan elimination routine.
 * Returns index of column which tested singular, or zero 
 * for success.  In case of error, the results are invalid and
 * the matrix is messed up, too.
 */
int gaussj(double *a, int n, double *b, int m)
{
   int *indxc, *indxr, *ipiv;
   int i, icol=0, irow=0, j, k, l, ll;
   double big, dum, pivinv;

   indxc = imalloc(n);
   indxr = imalloc(n);
   ipiv = imalloc(n);

   for (j=0; j<n; j++) ipiv[j]=0;
   for (i=0; i<n; i++)
   {
      big=0.0;
      for (j=0; j<n; j++)
         if (ipiv[j] != 1)
            for (k=0; k<n; k++)
            {
               if (ipiv[k] == 0)
               {
                  if (fabs(a[j*n + k]) >= big)
                  {
                     big=fabs(a[j*n + k]);
                     irow=j;
                     icol=k;
                  }
               }
               else if (ipiv[k] > 1)
               {
		 myfree(ipiv);
		 myfree(indxr);
		 myfree(indxc);
		 return k+1;
	       }
            }

      ++(ipiv[icol]);
      if (irow != icol)
      {
         for (l=0; l<n; l++) SWAP(a[irow*n + l],a[icol*n + l])
         for (l=0; l<m; l++) SWAP(b[irow*m + l],b[icol*m + l])
      }
      indxr[i]=irow;
      indxc[i]=icol;
      if (a[icol*n + icol] == 0.0)
      {
	myfree(ipiv);
	myfree(indxr);
	myfree(indxc);
	return icol+1;
      }
      pivinv = 1.0/a[icol*n + icol];
      a[icol*n + icol]=1.0;
      for (l=0; l<n; l++) a[icol*n + l] *= pivinv;
      for (l=0; l<m; l++) b[icol*m + l] *= pivinv;
      for (ll=0; ll<n; ll++)
         if (ll != icol)
         {
            dum=a[ll*n + icol];
            a[ll*n + icol]=0.0;
            for (l=0; l<n; l++) a[ll*n + l] -= a[icol*n + l]*dum;
            for (l=0; l<m; l++) b[ll*m + l] -= b[icol*m + l]*dum;
         }
   }
   for (l=n-1; l>=0; l--)
   {
      if (indxr[l] != indxc[l])
         for (k=0; k<n; k++)
            SWAP(a[k*n + indxr[l]],a[k*n + indxc[l]]);
   }


   myfree(ipiv);
   myfree(indxr);
   myfree(indxc);
   return 0;
}
#undef SWAP
