/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: mnbrak.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */
#include <math.h>
#include "tools.h"

#define GOLD 1.618034
#define GLIMIT 100 /* maximum parabolic magnification */

/* Bracket a minimum of a given function.
 *
 * Input:
 *   ax,bx = distinct parameters near the minimum (guess).
 *   fa,fb = function values at ax and bx.
 *   
 *   fun() = function to evaluate function and derivative.  It returns
 *		a non-zero error code if the function cannot be evaluated
 *		at the current value of x.  Also attached is an
 *		`environment' pointer which is simply passed.
 *
 * Output:
 *   ax,bx,cx = parameters below, near, and above the minimum.
 *   fa,fb,fc = function values at returned parameters
 *
 *   return value = 0 or non-zero code from fun() if something
 *   		terrible happened.
 *   
 * David Deaven deaven@ishmael.ameslab.gov (start w/numerical recipes)
 */
int mnbrak(double *ax, double *bx, double *cx,
  double *fa, double *fb, double *fc,
  int (*fun)(double x, double *y, double *dfun, void *fenv), void *fenv)
{
  int s;
  double ulim, q, r, dqr, u, fu, df;

  /* Ensure that fb < fa by swapping the two if not.
   */
  if (*fb > *fa)
  {
    double swap;
    swap = *ax; *ax = *bx; *bx = swap;
    swap = *fb; *fb = *fa; *fa = swap;
  }

  /* Get a valid third point to start the algorithm.
   */
  *cx = *bx + (*bx - *ax) * GOLD;
  if((s=(*fun)(*cx, fc, &df, fenv))) {
    /* Going further than bx is invalid.  Try choosing a point
     * between ax and bx, then give up if that also fails.
     */
    *cx = *bx;
    fc = fb;
    *bx = *ax + (*cx - *ax) / GOLD;
    if((s=(*fun)(*bx, fb, &df, fenv))) return s;
  }

  while (*fb >= *fc)
  {
    r = (*bx - *ax) * (*fb - *fc);
    q = (*bx - *cx) * (*fb - *fa);
    dqr = fabs(q - r);
    if(dqr < 1.e-20) dqr = 1.e-20;
    u = *bx - ((*bx - *cx) * q - (*bx - *ax) * r) /
      				 (copysign(dqr, q - r) * 2.0);
    ulim = *bx + (*cx - *bx) * GLIMIT;
    if ((*bx - u) * (u - *cx) > 0.0)
    {
	if((s=(*fun)(u, &fu, &df, fenv)))return s;
	if (fu < *fc) {
	    *ax = *bx;
	    *fa = *fb;
	    *bx = u;
	    *fb = fu;
	    continue;
	} else if (fu > *fb) {
	    *cx = u;
	    *fc = fu;
	    continue;
	}
	u = *cx + (*cx - *bx) * GOLD;
	if((s=(*fun)(u, &fu, &df, fenv)))return s;
    } else if ((*cx - u) * (u - ulim) > 0.0) {
	if((s=(*fun)(u, &fu, &df, fenv)))return s;
	if (fu < *fc) {
	    *bx = *cx;
	    *cx = u;
	    u = *cx + (*cx - *bx) * GOLD;
	    *fb = *fc;
	    *fc = fu;
	    if((s=(*fun)(u, &fu, &df, fenv)))return s;
	}
    } else if ((u - ulim) * (ulim - *cx) >= 0.0) {
	u = ulim;
	if((s=(*fun)(u, &fu, &df, fenv)))return s;
    } else {
	u = *cx + (*cx - *bx) * GOLD;
	if((s=(*fun)(u, &fu, &df, fenv)))return s;
    }
    *ax = *bx;
    *bx = *cx;
    *cx = u;
    *fa = *fb;
    *fb = *fc;
    *fc = fu;
    if(df == 0.0) return 0; /* flat spot */
  }
  return 0;
}
