/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: nordsk.c,v 1.4 1996/05/29 21:11:16 deaven Exp $
 * $Author: deaven $
 */

#include "memory.h"
#include "tools.h"

#define MAGIC_NUMBER	0xabeef

struct resources {
	int magic;	/* a safety feature */
	int n;		/* how many degrees of freedom */
	double *a1;
	double *a2;
};

/* Nordseick third-order variable timestep version; complete description
 * A. Nordseick, Math. Comput.(16)22(1962).
 *
 * Input:
 *   dimension = physical dimensions, used for accessing am[]
 *   n = number of coordinates
 *   hp = timestep
 *   x[] = coordinates
 *   v[] = velocities
 *   f[] = forces (also returned)
 *   e = potential energy
 *   am[] = masses, or NULL for all 1.0
 *   fun(double *x, *f, *e, int n) = routine to compute forces
 *
 * Output:
 *   f[] = newer forces
 *   hp = adjusted timestep (Don Brenner's method)
 *
 * David Deaven deaven@ishmael.ameslab.gov
 */
#include <stdio.h>
#include <math.h>
#include "memory.h"

/* For timestep adjustment, these values control the rate at which the
 * timestep responds to corrector errors.  ALPHA0 has length units,
 * and ALPHA1, EPSILON are dimensionless.
 */
#define ALPHA0 1.e-16	/* Don Brenner's values are 1.e-5 and 0.25 */
#define ALPHA1 0.25
#define EPSILON 1.e-5	/* Dave Deaven's addition */

void *new_nordsk(int n)
{
	struct resources *r = NewPointer(r);
	r->magic = MAGIC_NUMBER;
	r->n = n;
	r->a1 = dmalloc(n);
	r->a2 = dmalloc(n);
	return ((void *) r);
}

void delete_nordsk(void *res)
{
	struct resources *r = (struct resources *) res;
	if(r && r->magic == MAGIC_NUMBER) {
	  myfree(r->a1);
	  myfree(r->a2);
	  myfree(r);
	}
}

int nordsk(void *res, int dimension, int nc, double *hp,
	double x[], double v[], double f[], double *e, double *am,
	int (*fun)(double *, double *, double *, int))
{
  int i, n = dimension * nc;
  double h, h2, h3, q0, q1, q2, q3, ww, emax, err;
  double mi;
  struct resources *r = (struct resources *) res;

  if(r == NULL || r->magic != MAGIC_NUMBER || r->n != n) {
    dmd_errno = DMD_INVALID_ARGUMENT;
    return -1;
  }

  h = (*hp);
  h2=h*h*(1.0/2.0);
  h3=h2*h*(1.0/3.0);

  /* Predictor
   */
  for(i=0; i<n; ++i) {
    q0=x[i];
    q1=v[i]*h;
    q2=r->a1[i]*h2;
    q3=r->a2[i]*h3;
    x[i]=q0+q1+q2+q3;
    v[i]=q1+2.0*q2+3.0*q3;
    r->a1[i]=q2+3.0*q3;
    r->a2[i]=q3;
  }

  /* Evaluate(r)
   */
  if(fun(x, f, e, n)) {
    dmd_errno = DMD_COMPONENT_FAILURE;
    return -1;
  }

  /* Correcter
   */
  emax=0.0;
  for(i=0; i<n; ++i) { 
    mi = (am == NULL ? 1.0 : am[i / dimension]);
    ww=(f[i]*h2/mi-(r->a1[i])) * (1.0/6.0);
    x[i] += ww;
    v[i]=(v[i]+5.0*ww)/h;
    r->a1[i]=(r->a1[i]+6.0*ww)/h2;
    err = 2.0*ww;
    if(fabs(err) > emax) emax = fabs(err);
    r->a2[i]=(r->a2[i]+err)/h3;
  }

  /* Timestep adjustment
   */
  {
    double fact = pow(ALPHA0 / emax, ALPHA1);
    if(fact < (1.0 - EPSILON)) fact = (1.0 - EPSILON);
    if(fact > (1.0 + EPSILON)) fact = (1.0 + EPSILON);
    (*hp) = h * fact;
  }

  return 0;
}
