/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: rgb.c,v 1.5 1996/08/22 22:24:09 deaven Exp $
 * $Author: deaven $
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "memory.h"
#include "rgb.h"

static char *built_in[] = {
  "255   0   0   red",
  "  0 255   0   green",
  "  0   0 255   blue",
  "  0   0   0   black",
  "255 255 255   white",
  NULL
};

static struct {
  int ncache;			/* number of cached data */
  struct {
    char *name;
    Color color;
  } *cache;			/* cached color data */
  int npath;			/* number of searched files */
  char **path;			/* list of searched files */
} global = {0, NULL, 0, NULL};

/* Add a file name to the list of files used to store color
 * definitions.  This file will be searched and cached to resolve
 * colors not in the original table.
 */
void ColorAddDatabase(char *file)
{
	global.path = ResizePointerArray(global.path, global.npath + 1);
	global.path[global.npath++] = mystrdup(file);
}

static char *next_num(char *s, double *value)
{
	while(*s != '\0' && isspace(*s)) s++;
	if(sscanf(s, "%lf", value) != 1) return NULL;
	while(*s != '\0' && !isspace(*s)) s++; /* skip over the number */
	return s;
}

/* Read the RGB data and return the color name from
 * a single line of X11-style data.
 */
static char *grok_line(char *line, double rgb[3])
{
	while(*line != '\0' && isspace(*line)) line++;
	if(*line == '\0' || strchr("!#$;/", *line)) return ""; /* blank or comment line */

	if((line = next_num(line, rgb + 0)) == NULL) return "";
	if((line = next_num(line, rgb + 1)) == NULL) return "";
	if((line = next_num(line, rgb + 2)) == NULL) return "";

	while(*line != '\0' && isspace(*line)) line++;

	return line;
}

/* Add or replace the given definition in the cache.
 */
void ColorDefine(Color c, char *name)
{
	int i;
	Color *pc = NULL;

	for(i=0; i<global.ncache; i++) {
	  if(strcmp(name, global.cache[i].name) == 0) {
	    pc = &global.cache[i].color;
	    break;
	  }
	}

	if(pc == NULL) {
	  global.cache = ResizePointerArray(global.cache, global.ncache + 1);
	  global.cache[global.ncache].name = mystrdup(name);
	  pc = &(global.cache[global.ncache].color);
	  global.ncache++;
	}

	*pc = c;
}

/* Set the color equal to the value encoded in "name", which may be
 * either one of the pre-defined names, a hard specification, or
 * a value read from the X11 color database or equivalent.
 */
Color ColorFromName(char *name)
{
	int i, r, g, b;
	double rgb[3];
	Color c;

	/* Check for a cached definition.
	 */
	for(i=0; i<global.ncache; i++) {
	  if(strcmp(name, global.cache[i].name) == 0)
	    return global.cache[i].color;
	}

	/* Scan the built-in names next.
	 */
	for(i=0; built_in[i]; i++) {
	  if(strcmp(grok_line(built_in[i], rgb), name) == 0) {
	    c.r = rgb[0];
	    c.g = rgb[1];
	    c.b = rgb[2];
	    ColorDefine(c, name); /* cache for next time */
	    return c;
	  }
	}

	/* Next, try to read an RGB color in the 
	 * format "rgb:r/g/b" with each number in hex.
	 */
        if(strncmp(name, "rgb:", 4) == 0) {
	  if(sscanf(name + 4, "%x/%x/%x", &r, &g, &b) == 3) {
	    c.r = r;
	    c.g = g;
	    c.b = b;
	    return c;
	  }
        }

	/* Finally, read the search files and see if it is defined there.
	 */
	{ FILE *f;
	  char buf[128];
	  for(i=0; i<global.npath; i++) {
	    if((f = fopen(global.path[i], "r")) == NULL) continue;
	    while(fgets(buf, 128, f)) {
	      if(strncmp(grok_line(buf, rgb), name, strlen(name)) == 0) {
		c.r = rgb[0] * ((double) RGB_CMAX / 255.0);
		c.g = rgb[1] * ((double) RGB_CMAX / 255.0);
		c.b = rgb[2] * ((double) RGB_CMAX / 255.0);
		fclose(f);
		ColorDefine(c, name); /* cache for next time */
		return c;
	      }
	    }
	    fclose(f);
	  }
	}

        /* Last, just return the default color: black.
         */
	c.r = c.g = c.b = 0;
	return c;
}

/* Given a color, return a string that setrgbcolor() can convert.
 */
char *ColorName(Color c)
{
	int i;
	static char buf[64];

	/* First try to read a defined name.
	 */
	for(i=0; i<global.ncache; i++) {
	    if(c.r == global.cache[i].color.r &&
	       c.g == global.cache[i].color.g &&
	       c.b == global.cache[i].color.b) return global.cache[i].name;
	}

	/* If that fails, construct an explicit specification.
	 */
	sprintf(buf, "rgb:%x/%x/%x", c.r, c.g, c.b);
	return buf;
}
