/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: slaterkoster.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

#include "slaterkoster.h"

/* Index and sign tables into packed Slater-Koster (SK) arrays.
 *
 * Indices from 1-10 label combined orbital indices,
 * {s.s s.px s.py s.pz px.px px.py px.pz py.py py.pz pz.pz}
 *
 * A(ij) = SKstab[i][j] * Apacked(SKitab[i][j])
 *
 * Such an orbital matrix is stored in the following order, with 
 * signs to reflect SKstab:
 *
 *			 0  1  2  3  
 *			-1  4  5  6
 *	element (02)--> -2  5  7  8
 *			-3  6  8  9
 *
 */
int SKitab[NORBAT][NORBAT]={
  {0, 1, 2, 3}, {1, 4, 5, 6}, {2, 5, 7, 8}, {3, 6, 8, 9} };
int SKstab[NORBAT][NORBAT]={
  {1, 1, 1, 1}, {-1, 1, 1, 1}, {-1, 1, 1, 1}, {-1, 1, 1, 1} };

/*
 * get Slater Koster matrix.
 * Compute NORBAT x NORBAT matrix from bare constants
 *
 * INPUT:
 * sp() = {ss0,sp0,pp0,pp1}
 * c[xyz] = (x_j-x_i)/d(ij)
 *
 * OUTPUT:
 * s() = i-->j Slater-Koster matrix (returned)
 *
 * Swapping orbital indices gives a minus sign in the sp0 elements,
 * as does swapping site indices (c[xyz] --> -c[xyz]).  Thus 10 
 * elements are all that's needed.
 *
 * David Deaven 1 February 1994 <deaven@ishmael.ameslab.gov>
 *
 */
void
SKget(double sp[4], double cx, double cy, double cz, double s[10])
{
  double a=sp[2]-sp[3];
  s[0]=sp[0];
  s[1]=sp[1]*cx;
  s[2]=sp[1]*cy;
  s[3]=sp[1]*cz;
  s[4]=cx*cx*a+sp[3];
  s[5]=cx*cy*a;
  s[6]=cx*cz*a;
  s[7]=cy*cy*a+sp[3];
  s[8]=cy*cz*a;
  s[9]=cz*cz*a+sp[3];
}

/*
 * get Derivative of Slater-Koster geometrical factors.
 *
 * INPUT:
 * sh(4) = {ss0 sp0 pp0 pp1} off-site matrix elements
 * c{x,y,z} = |x(j)-x(i)|/d(ij) 
 *
 * OUTPUT:
 * ds(10,3) = d(ij)*derivatives of matrix elements wrt x_j or -x_i.
 *
 * We incur the same sign changes as in getSK() wrt swapping orbital
 * indices, and the sign change for site index swap is all _except_ the
 * sp0 elements; thus the result of a combined orbital and site index
 * swap is an overall minus sign.
 *
 * David Deaven 31 January 1994 <deaven@ishmael.ameslab.gov>
 *
 */
void 
SKgetd(double sh[4], double cx, double cy, double cz, double ds[3][10])
{
  double a=sh[2]-sh[3];

  ds[0][0]=0.0;
  ds[1][0]=0.0;
  ds[2][0]=0.0;

  ds[0][1]=sh[1]*(1.0-cx*cx);
  ds[1][1]=sh[1]*(-cy*cx);
  ds[2][1]=sh[1]*(-cz*cx);

  ds[0][2]=sh[1]*(-cx*cy);
  ds[1][2]=sh[1]*(1.0-cy*cy);
  ds[2][2]=sh[1]*(-cz*cy);

  ds[0][3]=sh[1]*(-cx*cz);
  ds[1][3]=sh[1]*(-cy*cz);
  ds[2][3]=sh[1]*(1.0-cz*cz);

  ds[0][4]=2.0*cx*(1.0-cx*cx)*a;
  ds[1][4]=-2.0*cy*cx*cx*a;
  ds[2][4]=-2.0*cz*cx*cx*a;

  ds[0][5]=cy*(1.0-2.0*cx*cx)*a;
  ds[1][5]=cx*(1.0-2.0*cy*cy)*a;
  ds[2][5]=-2.0*cx*cy*cz*a;

  ds[0][6]=cz*(1.0-2.0*cx*cx)*a;
  ds[1][6]=-2.0*cx*cy*cz*a;
  ds[2][6]=cx*(1.0-2.0*cz*cz)*a;

  ds[0][7]=-2.0*cx*cy*cy*a;
  ds[1][7]=2.0*cy*(1.0-cy*cy)*a;
  ds[2][7]=-2.0*cz*cy*cy*a;

  ds[0][8]=-2.0*cx*cy*cz*a;
  ds[1][8]=cz*(1.0-2.0*cy*cy)*a;
  ds[2][8]=cy*(1.0-2.0*cz*cz)*a;

  ds[0][9]=-2.0*cx*cz*cz*a;
  ds[1][9]=-2.0*cy*cz*cz*a;
  ds[2][9]=2.0*cz*(1.0-cz*cz)*a;
}

/* 
 * Unpack a packed Slater-Koster matrix
 * c(lm,l'm')=a(lml'm')
 *
 * David Deaven 2 February 1994 <deaven@ishmael.ameslab.gov>
 *
 */
void SKunpack(double a[10], double c[NORBAT][NORBAT])
{
  c[0][0]=a[0];
  c[0][1]=-a[1];
  c[0][2]=-a[2];
  c[0][3]=-a[3];
  c[1][1]=a[4];
  c[1][2]=a[5];
  c[1][3]=a[6];
  c[2][2]=a[7];
  c[2][3]=a[8];
  c[3][3]=a[9];
  c[1][0]=a[1];
  c[2][0]=a[2];
  c[3][0]=a[3];
  c[2][1]=a[5];
  c[3][1]=a[6];
  c[3][2]=a[8];
}

