/* Routines for constructing cubic splines
 * This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: spline.c,v 1.2 1995/08/29 22:45:15 deaven Exp $
 * $Author: deaven $
 */

#include <math.h>
#include <stdlib.h>
#include "spline.h"
#include "memory.h"

#define ZEPS 1.e-100 /* zero epsilon */

Spline1D *newSpline1D(int n)
{
	Spline1D *new = mymalloc(sizeof(Spline1D));
	new->n = n;
	if(n > 0)
	{
	  new->x = dmalloc(n);
	  new->y = dmalloc(n);
	  new->d = dmalloc(n);
	}
	else
	  new->x = new->y = new->d = NULL;

	return new;
}

Spline1D *resizeSpline1D(Spline1D *s, int n)
{
	if(s == NULL) return newSpline1D(n);

	s->x = myrealloc(s->x, n * sizeof(double));
	s->y = myrealloc(s->y, n * sizeof(double));
	s->d = myrealloc(s->d, n * sizeof(double));
	s->n = n;
	return s;
}

void freeSpline1D(Spline1D *old)
{
	if(old != NULL) {
	  myfree(old->x);
	  myfree(old->y);
	  myfree(old->d);
	  myfree(old);
	}
}

/*==========================================================================
 * David Deaven 23 February 1994 deaven@ishmael.ameslab.gov 
 *========================================================================*/
void spline(Spline1D *s, double yp1, double ypn)
{
  double *u, sig, p, qn, un;
  double *x=s->x, *y=s->y, *d=s->d;
  double dxa, dxb;
  int i, n=s->n;
  u=dmalloc(n);
  if (yp1 == SPL_NATURAL) {
    d[0]=0.0;
    u[0]=0.0;
  } else {
    if (yp1 == SPL_LINEAR)
      yp1 = (y[1] - y[0]) / (x[1] - x[0]);
    d[0]=-0.5;
    u[0]=(3.0/(x[1]-x[0]))*((y[1]-y[0])/(x[1]-x[0])-yp1);
  }
#if 0
  for(i=1; i<(n-1); ++i) {
    sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
    p=sig*d[i-1]+2.0;
    d[i]=(sig-1.0)/p;
    u[i]=(6.0*((y[i+1]-y[i])/(x[i+1]-x[i]) - 
	    (y[i]-y[i-1])/(x[i]-x[i-1]))/(x[i+1]-x[i-1])-sig*u[i-1])/p;
  }
#else
  for(i=1; i<(n-1); ++i)
  {
    dxa = x[i]-x[i-1];
    dxb = x[i+1]-x[i];
    if(fabs(dxb) < ZEPS || fabs(dxa) < ZEPS)
    {
      fprintf(stderr, "spline: constant x value %g (near element %d)\n",
	x[i], i+1);
      if(fabs(dxa) < ZEPS) dxa = ZEPS;
      if(fabs(dxb) < ZEPS) dxb = ZEPS;
    }
    sig = dxa / (dxa + dxb);
    p = 1.0 / (sig * d[i-1] + 2.0);
    d[i] = p * (sig - 1.0);
    u[i] = p * (6.0 * ((y[i+1]-y[i])/dxb - 
		       (y[i]-y[i-1])/dxa   ) / (dxa + dxb) - sig * u[i-1]);
  }
#endif
  if (ypn == SPL_NATURAL) {
    qn=0.0;
    un=0.0;
  } else {
    if (ypn == SPL_LINEAR)
      ypn = (y[n-1] - y[n-2]) / (x[n-1] - x[n-2]);
    qn=0.5;
    un=(3.0/(x[n-1]-x[n-2]))*(ypn-(y[n-1]-y[n-2])/(x[n-1]-x[n-2]));
  }
  d[n-1]=(un-qn*u[n-2])/(qn*d[n-2]+1.0);
  for(i=n-2; i>=0; --i) d[i] = d[i]*d[i+1]+u[i];

  myfree(u);
}

