/* This file is part of David Deaven's personal library.
 * Copyright (c) 1994 David Deaven
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: ypr.c,v 1.5 1996/06/27 17:07:04 deaven Exp $
 * $Author: deaven $
 */

#include <stdio.h>
#include <math.h>
#include "tools.h"

/* Generate a rotational transformation using the common aerodynamic
 * angles yaw, pitch, and roll.  From Goldstein's point of view (as
 * given in _Classical_Mechanics_, appendix B) we are doing the "321"
 * convention, and computing the rotation matrix which when applied to
 * a body frame point gives the space frame point.  This is exactly
 * the transpose of the matrix found in appendix B, Eq.(B-11xyz),
 * which rotates the body frame _axes_ into the physical axes, and
 * with negated angles.
 *
 * yaw around z, pitch around local y, roll around final x.
 */
double *YPRRotationMatrix(double yaw, double pitch, double roll, double t[3*3])
{
	double s1 = sin(yaw),   c1 = cos(yaw);
	double s2 = sin(pitch), c2 = cos(pitch);
	double s3 = sin(roll),  c3 = cos(roll);

	t[0] = c1 * c2;
	t[1] = s1 * c3 + c1 * s2 * s3;
	t[2] = s1 * s3 - c1 * s2 * c3;

	t[3] = -s1 * c2;
	t[4] = c1 * c3 - s1 * s2 * s3;
	t[5] = c1 * s3 + s1 * s2 * c3;

	t[6] = s2;
	t[7] = -c2 * s3;
	t[8] = c2 * c3;

	return t;
}

/* Given for n rigid rotors:
 *    ypr[3*n], the YPR orientation of each rotor,
 *    w[3*n], the body frame angular velocity of each rotor,
 *    I[3*n], the principal moments of inertia for each rotor,
 *    torque[3*n], the space frame torques on each rotor,
 *    h, a timestep;
 * Advance the YPR orientations ypr[] and body-frame angular
 * velocities w[] given the torque[], and return the rotational
 * kinetic energy.
 *   
 * David Deaven, January 1996.
 */
double YPRIntegrate(int n, double *ypr, double *w, double *I,
	double h, double *torque)
{
	int i;
	double *ai, *wi, *ti, *Ii, dai[3], dwi[3];
	double b[3*3], tis[3];
	double s2, c2, s1, c1;
	double ek = 0.0;
	static double Iunit[3] = {1.0, 1.0, 1.0};

	for(i=0; i<n; ++i)
	{
	  ai = ypr + 3*i;
	  wi = w + 3*i;
	  ti = torque + 3*i;
	  Ii = I ? I + 3*i : Iunit;

	  /* These are equations of motion in the body frame.
	   * The dwi[] are the _body_frame_ angular accelerations.
	   * We have to convert the space frame torques (ti[]) to the
	   * body frame (tis[]) first.
	   */
	  mmul3(YPRRotationMatrix(ai[0], ai[1], ai[2], b), ti, tis);
	  dwi[0] = (-tis[0] + wi[1] * wi[2] * (Ii[1] - Ii[2])) / Ii[0];
	  dwi[1] = (-tis[1] + wi[2] * wi[0] * (Ii[2] - Ii[0])) / Ii[1];
	  dwi[2] = (-tis[2] + wi[0] * wi[1] * (Ii[0] - Ii[1])) / Ii[2];

	  /* This is the transformation to the ypr orientation
	   * angles (ai[]) and derivatives (dai[]) from the _body_frame_ 
	   * angular velocities (wi[]).
	   */
	  s2 = -sin(ai[2]); c2 = cos(ai[2]);
	  s1 = -sin(ai[1]); c1 = cos(ai[1]);

	  dai[0] = (wi[2] * c2 + wi[1] * s2) / c1;
	  dai[1] = wi[1] * c2 - wi[2] * s2;
	  dai[2] = wi[0] + s1 * dai[0];

	  ai[0] += h * dai[0];
	  ai[1] += h * dai[1];
	  ai[2] += h * dai[2];

	  wi[0] += h * dwi[0];
	  wi[1] += h * dwi[1];
	  wi[2] += h * dwi[2];

	  ek += 0.5 * (Ii[0] * wi[0] * wi[0] +
		       Ii[1] * wi[1] * wi[1] +
		       Ii[2] * wi[2] * wi[2]   );
	}

	return ek;
}

/* Compute the transformation matrix to convert a derivative wrt. the
 * space axes into a derivative wrt. the YPR angles.
 *
 * d/d(YPR) = T d/d(theta-space)
 *
 * These come directly from Eq.(B-15xyz) in Goldstein Appendix B,
 * except that we have negated YPR angles compared to him, and we
 * want the *transpose* of the matrix he gives.
 */
double *YPRTransformDerivative(double xi[3], double t[3*3])
{
        double s1 = -sin(xi[0]), c1 = cos(xi[0]);
        double s2 = -sin(xi[1]), c2 = cos(xi[1]);

        t[0] = 0.0;
        t[3] = -s1;
        t[6] = c2 * c1;

        t[1] = 0.0;
        t[4] = c1;
        t[7] = c2 * s1;

        t[2] = 1.0;
        t[5] = 0.0;
        t[8] = -s2;

	return t;
}

/* Convert YPR orientation angles into quaternions.
 * This follows from Goldstein Eq.(B-13xyz) except with negated YPR
 * angles since this code rotates coordinates instead of axes.
 */
double *YPRtoQ(double ypr[3], double q[4])
{
	int i;
	double c[3], s[3];

	for(i=0; i<3; i++) {
	  c[i] = cos(0.5 * ypr[i]);
	  s[i] = sin(0.5 * ypr[i]);
	}

	q[0] =  c[0]*c[1]*c[2] - s[0]*s[1]*s[2];
	q[1] = -c[0]*c[1]*s[2] - s[0]*s[1]*c[2];
	q[2] = -c[0]*s[1]*c[2] + s[0]*c[1]*s[2];
	q[3] = -c[0]*s[1]*s[2] - s[0]*c[1]*c[2];

	return q;
}

static void rot_y(double theta, double x[3])
{
	double c = cos(theta), s = sin(theta), xx[2];
	xx[0] = c * x[0] - s * x[2];
	xx[1] = s * x[0] + c * x[2];
	x[0] = xx[0];
	x[2] = xx[1];
}

static void rot_z(double theta, double x[3])
{
	double c = cos(theta), s = sin(theta), xx[2];
	xx[0] = c * x[0] - s * x[1];
	xx[1] = s * x[0] + c * x[1];
	x[0] = xx[0];
	x[1] = xx[1];
}

/* Given a rotation matrix R, determine the YPR orientation angles
 * needed to generate the same transformation.
 */
double *YPRfromR(double R[3*3], double ypr[3])
{
	int i;
	double x[3], y[3], z[3];

	for(i=0; i<3; i++) {
	  x[i] = R[i*3];
	  y[i] = R[i*3 + 1];
	  z[i] = R[i*3 + 2];
	}

	/* I. Apply a yaw to get the correct x direction in the xy plane.
	 *    (Opposite of rotating x[] to make x[1] = 0.)
	 */
	ypr[0] = -atan2(x[1], x[0]);
	rot_z(ypr[0], x);
	rot_z(ypr[0], z);

	/* II. Apply a pitch to get the correct x direction in the
	 *     xz plane. (Opposite of making x[2] = 0.)
	 */
	ypr[1] = atan2(x[2], x[0]);
	rot_y(-ypr[1], z);

	/* III. Apply a roll to bring z[] to the z axis.
	 *      (Opposite of making z[1] = 0.)
	 */
	ypr[2] = atan2(z[1], z[2]);

	return ypr;
}

